from pytest import fixture, raises

from pddl_plus_parser.lisp_parsers import DomainParser, PDDLTokenizer
from pddl_plus_parser.models import PDDLType, Predicate, Action
from tests.lisp_parsers_tests.consts import TEST_PARSING_FILE_PATH, TEST_WOODWORKING_DOMAIN_PATH, \
    TEST_NUMERIC_DEPOT_DOMAIN_PATH, PLANT_WATERING_DOMAIN, TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN, \
    TEST_TYPES_FOR_CONDITIONAL_DOMAIN, TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN, SPIDER_DOMAIN_PATH, \
    TYPES_FOR_UNIVERSAL_CONDITIONAL_DOMAIN, TEST_PREDICATES_FOR_UNIVERSAL_QUANTIFIER_DOMAIN

test_types_with_no_parent = ['acolour', 'awood', 'woodobj', 'machine', 'surface', 'treatmentstatus', 'aboardsize',
                             'apartsize']

nested_types = ['acolour', 'awood', 'woodobj', 'machine', 'surface', 'treatmentstatus', 'aboardsize', 'apartsize',
                '-', 'object', 'highspeed-saw', 'saw', 'glazer', 'grinder', 'immersion-varnisher', 'planer',
                'spray-varnisher', '-', 'machine', 'board', 'part', '-', 'woodobj']

test_constants = ['small', 'medium', 'large', '-', 'apartsize', 'highspeed-saw', 'varnished', 'glazed', 'untreated',
                  'colourfragments', '-', 'treatmentstatus',
                  'natural', '-', 'acolour', 'verysmooth', 'smooth', 'rough', '-', 'surface']

test_predicates_str = """((available ?obj - woodobj)
	(surface-condition ?obj - woodobj ?surface - surface)
	(treatment ?obj - part ?treatment - treatmentstatus)
	(colour ?obj - part ?colour - acolour)
	(wood ?obj - woodobj ?wood - awood)
	(is-smooth ?surface - surface)
	(has-colour ?agent - machine ?colour - acolour)
	(goalsize ?part - part ?size - apartsize)
	(boardsize-successor ?size1 - aboardsize ?size2 - aboardsize)
	(unused ?obj - part)
	(boardsize ?board - board ?size - aboardsize)
	(empty ?agent - highspeed-saw)
	(in-highspeed-saw ?b - board ?agent - highspeed-saw)
	(grind-treatment-change ?agent - grinder ?old - treatmentstatus ?new - treatmentstatus))"""


@fixture()
def domain_parser() -> DomainParser:
    return DomainParser(TEST_PARSING_FILE_PATH)


def test_parse_types_with_no_parent_extracts_types_with_object_as_parent(domain_parser: DomainParser):
    parsed_types = domain_parser.parse_types(test_types_with_no_parent)
    assert len(parsed_types) == len(test_types_with_no_parent) + 1
    for type_name, expected_type_name in zip(parsed_types.keys(), test_types_with_no_parent):
        assert type_name == expected_type_name
        assert parsed_types[type_name].parent.name == "object"


def test_parse_types_with_object_parent_not_create_duplicates(domain_parser: DomainParser):
    test_types_with_object_parent = ['acolour', 'awood', 'woodobj', 'machine', 'surface', 'treatmentstatus',
                                     'aboardsize', 'apartsize', '-', 'object']

    parsed_types = domain_parser.parse_types(test_types_with_object_parent)
    assert len(parsed_types) == len(test_types_with_no_parent) + 1
    for type_name, expected_type_name in zip(parsed_types.keys(), test_types_with_no_parent):
        assert type_name == expected_type_name
        assert parsed_types[type_name].parent.name == "object"


def test_parse_types_with_deep_type_hierarchy_recognizes_ancestors_correctly(domain_parser: DomainParser):
    test_types_deep_hierarchy = ['place', 'locatable', '-', 'object', 'hoist', 'surface',
                                 '-', 'locatable', 'pallet', '-', 'surface']

    parsed_types = domain_parser.parse_types(test_types_deep_hierarchy)
    assert "pallet" in parsed_types
    assert parsed_types["pallet"].parent.name == "surface"
    assert parsed_types["pallet"].parent.parent.name == "locatable"
    assert parsed_types["pallet"].parent.parent.parent.name == "object"


def test_parse_types_with_type_hierarchy_recognize_nested_types(domain_parser: DomainParser):
    parsed_types = domain_parser.parse_types(nested_types)
    for object_descendant in test_types_with_no_parent:
        assert parsed_types[object_descendant].parent.name == "object"

    machine_types = ['highspeed-saw', 'saw', 'glazer', 'grinder', 'immersion-varnisher', 'planer', 'spray-varnisher']
    for machine_descendant in machine_types:
        assert parsed_types[machine_descendant].parent.name == "machine"

    wood_object_types = ['board', 'part']
    for wood_obj_descendant in wood_object_types:
        assert parsed_types[wood_obj_descendant].parent.name == "woodobj"

    assert parsed_types["machine"].parent.name == "object"
    assert parsed_types["woodobj"].parent.name == "object"


def test_parse_constants_when_given_invalid_type_raises_syntax_error(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    with raises(SyntaxError):
        bad_constants = ['small', 'medium', 'large', '-', 'bad-type']
        domain_parser.parse_constants(bad_constants, domain_types)


def test_parse_constants_when_given_valid_type_extract_constants(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    valid_constants = ['small', 'medium', 'large', '-', 'apartsize']
    constants = domain_parser.parse_constants(valid_constants, domain_types)
    assert len(constants) == 3
    for const in constants.values():
        assert const.name in valid_constants


def test_parse_constants_with_nexted_constants_extract_correct_constants_data(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    domain_consts = domain_parser.parse_constants(test_constants, domain_types)

    assert list(domain_consts.keys()) == ['small', 'medium', 'large', 'highspeed-saw', 'varnished', 'glazed',
                                          'untreated', 'colourfragments', 'natural', 'verysmooth', 'smooth', 'rough']


def test_parse_predicate_with_legal_predicate_data_is_successful(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    test_predicate = ['available', '?obj', '-', 'woodobj']
    predicate = domain_parser.parse_predicate(test_predicate, domain_types)
    assert predicate.name == "available"
    assert "?obj" in predicate.signature


def test_parse_predicate_with_illegal_predicate_data_raises_error(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    with raises(SyntaxError):
        test_predicate = ['available', '?obj', 'woodobj']
        domain_parser.parse_predicate(test_predicate, domain_types)


def test_parse_predicate_with_no_parameters_returns_predicate_object_with_only_name_and_empty_signature(
        domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    test_predicate_no_params_ast = ["ok"]
    predicate = domain_parser.parse_predicate(test_predicate_no_params_ast, domain_types)
    assert predicate is not None
    assert predicate.name == "ok"
    assert len(predicate.signature) == 0


def test_parse_predicates_with_single_predicate_returns_predicates_dictionary_correctly(domain_parser: DomainParser):
    domain_types = domain_parser.parse_types(nested_types)
    test_predicates = [['available', '?obj', '-', 'woodobj']]
    predicates = domain_parser.parse_predicates(test_predicates, domain_types)
    assert "available" in predicates


def test_parse_predicates_with_private_predicates_extracts_predicates_correctly(domain_parser: DomainParser):
    private_predicates_str = """((:private
	(surface-condition ?obj - woodobj ?surface - surface)
	(treatment ?obj - part ?treatment - treatmentstatus)
    ))"""

    tokenizer = PDDLTokenizer(pddl_str=private_predicates_str)
    domain_types = domain_parser.parse_types(nested_types)
    predicates = domain_parser.parse_predicates(tokenizer.parse(), domain_types)
    assert "surface-condition" in predicates
    assert predicates["surface-condition"].untyped_representation == "(surface-condition ?obj ?surface)"
    assert "treatment" in predicates
    assert predicates["treatment"].untyped_representation == "(treatment ?obj ?treatment)"


def test_parse_predicates_with_multiple_predicate_returns_predicates_dictionary_correctly(domain_parser: DomainParser):
    tokenizer = PDDLTokenizer(pddl_str=test_predicates_str)
    domain_types = domain_parser.parse_types(nested_types)
    predicates = domain_parser.parse_predicates(tokenizer.parse(), domain_types)
    assert len(predicates) == 14


def test_parse_functions_with_single_function_extract_function_data_correctly(domain_parser: DomainParser):
    test_pddl_function = "((velocity ?saw - highspeed-saw))"
    tokenizer = PDDLTokenizer(pddl_str=test_pddl_function)
    domain_types = domain_parser.parse_types(nested_types)
    functions = domain_parser.parse_functions(tokenizer.parse(), domain_types)
    assert len(functions) == 1
    assert "velocity" in functions
    assert str(functions["velocity"]) == "(velocity ?saw - highspeed-saw)"


def test_parse_functions_with_multiple_functions_extract_functions_correctly(domain_parser: DomainParser):
    test_pddl_functions = "((velocity ?saw - highspeed-saw)" \
                          "(distance-to-floor ?m - machine))"
    tokenizer = PDDLTokenizer(pddl_str=test_pddl_functions)
    domain_types = domain_parser.parse_types(nested_types)
    functions = domain_parser.parse_functions(tokenizer.parse(), domain_types)
    assert len(functions) == 2


def test_parse_function_with_no_parameters_returns_function_object_with_only_name_and_empty_signature(
        domain_parser: DomainParser):
    test_function_no_params = "((total-cost))"
    tokenizer = PDDLTokenizer(pddl_str=test_function_no_params)
    domain_types = domain_parser.parse_types(nested_types)
    functions = domain_parser.parse_functions(tokenizer.parse(), domain_types)
    assert functions is not None
    assert len(functions) == 1
    assert "total-cost" in functions


def test_parse_action_with_inequality_precondition_adds_the_lifted_objects_correctly(domain_parser: DomainParser):
    test_action_str = """(cut-board-medium
	:parameters   (?m - highspeed-saw ?b - board ?p - part ?w - awood ?surface - surface ?size_before - aboardsize ?s1 - aboardsize ?size_after - aboardsize)
	:precondition (and (unused ?p) (not (= ?size_before ?size_after)) (goalsize ?p medium) (in-highspeed-saw ?b ?m) (wood ?b ?w) (surface-condition ?b ?surface) (boardsize ?b ?size_before) (boardsize-successor ?size_after ?s1) (boardsize-successor ?s1 ?size_before))
	:effect       (and (boardsize ?b ?size_after) (available ?p) (surface-condition ?p ?surface) (treatment ?p untreated) (wood ?p ?w) (colour ?p natural) (not (unused ?p))))"""
    action_tokens = PDDLTokenizer(pddl_str=test_action_str).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=test_predicates_str).parse()
    domain_types = domain_parser.parse_types(nested_types)
    domain_consts = domain_parser.parse_constants(test_constants, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    action = domain_parser.parse_action(action_tokens, domain_types, domain_functions, domain_predicates, domain_consts)

    assert action.name == "cut-board-medium"
    assert action.inequality_preconditions == {("?size_before", "?size_after")}


def test_parse_action_with_equality_precondition_adds_the_lifted_objects_correctly(domain_parser: DomainParser):
    test_action_str = """(cut-board-medium
	:parameters   (?m - highspeed-saw ?b - board ?p - part ?w - awood ?surface - surface ?size_before - aboardsize ?s1 - aboardsize ?size_after - aboardsize)
	:precondition (and (unused ?p) (= ?size_before ?size_after) (goalsize ?p medium) (in-highspeed-saw ?b ?m) (wood ?b ?w) (surface-condition ?b ?surface) (boardsize ?b ?size_before) (boardsize-successor ?size_after ?s1) (boardsize-successor ?s1 ?size_before))
	:effect       (and (boardsize ?b ?size_after) (available ?p) (surface-condition ?p ?surface) (treatment ?p untreated) (wood ?p ?w) (colour ?p natural) (not (unused ?p))))"""
    action_tokens = PDDLTokenizer(pddl_str=test_action_str).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=test_predicates_str).parse()
    domain_types = domain_parser.parse_types(nested_types)
    domain_consts = domain_parser.parse_constants(test_constants, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    action = domain_parser.parse_action(action_tokens, domain_types, domain_functions, domain_predicates, domain_consts)

    assert action.name == "cut-board-medium"
    assert action.equality_preconditions == {("?size_before", "?size_after")}


def test_parse_action_with_boolean_action_type_returns_action_data_correctly(domain_parser: DomainParser):
    test_action_str = """(do-spray-varnish
	:parameters   (?m - spray-varnisher ?x - part ?newcolour - acolour ?surface - surface)
	:precondition (and (available ?x) (has-colour ?m ?newcolour) (surface-condition ?x ?surface) (is-smooth ?surface) (treatment ?x untreated))
	:effect       (and (treatment ?x varnished) (colour ?x ?newcolour) (not (treatment ?x untreated)) (not (colour ?x natural))))"""
    action_tokens = PDDLTokenizer(pddl_str=test_action_str).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=test_predicates_str).parse()
    domain_types = domain_parser.parse_types(nested_types)
    domain_consts = domain_parser.parse_constants(test_constants, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    action = domain_parser.parse_action(action_tokens, domain_types, domain_functions, domain_predicates, domain_consts)

    assert action.name == "do-spray-varnish"
    assert action.signature == {
        "?m": PDDLType("spray-varnisher"),
        "?x": PDDLType("part"),
        "?newcolour": PDDLType("acolour"),
        "?surface": PDDLType("surface"),
    }
    assert action.negative_preconditions == set()
    expected_preconditions = {
        Predicate(name="available", signature={"?x": PDDLType("part")}),
        Predicate(name="has-colour", signature={
            "?m": PDDLType("spray-varnisher"),
            "?newcolour": PDDLType("acolour")
        }),
        Predicate(name="surface-condition", signature={
            "?x": PDDLType("part"),
            "?surface": PDDLType("surface")
        }),
        Predicate(name="is-smooth", signature={
            "?surface": PDDLType("surface")
        }),
        Predicate(name="treatment", signature={
            "?x": PDDLType("part"),
            "untreated": PDDLType("treatmentstatus")
        }),

    }
    assert len(action.positive_preconditions) == len(expected_preconditions)
    for precondition in expected_preconditions:
        assert precondition in action.positive_preconditions

    expected_add_effects = {
        Predicate(name="colour", signature={
            "?x": PDDLType("part"),
            "?newcolour": PDDLType("acolour")
        }),
        Predicate(name="treatment", signature={
            "?x": PDDLType("part"),
            "varnished": PDDLType("treatmentstatus")
        })
    }
    assert len(action.add_effects) == len(expected_add_effects)
    for effect in expected_add_effects:
        assert effect in action.add_effects

    expected_delete_effects = {
        Predicate(name="treatment", signature={
            "?x": PDDLType("part"),
            "untreated": PDDLType("treatmentstatus")
        }),
        Predicate(name="colour", signature={
            "?x": PDDLType("part"),
            "natural": PDDLType("acolour")
        })
    }

    assert len(action.add_effects) == len(expected_delete_effects)
    for effect in expected_delete_effects:
        assert effect in action.delete_effects


def test_parse_effects_with_conditional_effects_with_one_condition_and_one_effect_parse_correctly(
        domain_parser: DomainParser):
    conditional_effects = """(and (when
            (not (CAN-CONTINUE-GROUP ?c ?to))
            (make-unmovable ?to)
    ))"""
    types_tokens = PDDLTokenizer(pddl_str=TEST_TYPES_FOR_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN).parse()
    constants_tokens = PDDLTokenizer(pddl_str=TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN).parse()
    effects_tokens = PDDLTokenizer(pddl_str=conditional_effects).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_consts = domain_parser.parse_constants(constants_tokens, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    new_action = Action()
    new_action.name = "test-action"
    new_action.signature = {"?c": domain_types["card"], "?from": domain_types["cardposition"],
                            "?fromdeal": domain_types["deal"], "?to": domain_types["card"],
                            "?totableau": domain_types["tableau"]}
    domain_parser.parse_effects(effects_ast=effects_tokens,
                                new_action=new_action,
                                domain_types=domain_types,
                                domain_functions=domain_functions,
                                domain_predicates=domain_predicates,
                                domain_constants=domain_consts)
    conditional_effect = new_action.conditional_effects.pop()
    assert conditional_effect is not None
    print(str(conditional_effect))
    assert str(
        conditional_effect) == "(when (and (not (CAN-CONTINUE-GROUP ?c ?to))) (and (make-unmovable ?to)))".lower()


def test_parse_effects_with_conditional_effects_with_one_condition_and_two_effects_parse_correctly(
        domain_parser: DomainParser):
    conditional_effects = """(and (when
            (not (can-continue-group ?c ?to))
            (and
                (currently-updating-unmovable)
                (make-unmovable ?to)
            )
    ))"""
    types_tokens = PDDLTokenizer(pddl_str=TEST_TYPES_FOR_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN).parse()
    constants_tokens = PDDLTokenizer(pddl_str=TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN).parse()
    effects_tokens = PDDLTokenizer(pddl_str=conditional_effects).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_consts = domain_parser.parse_constants(constants_tokens, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    new_action = Action()
    new_action.name = "test-action"
    new_action.signature = {"?c": domain_types["card"], "?from": domain_types["cardposition"],
                            "?fromdeal": domain_types["deal"], "?to": domain_types["card"],
                            "?totableau": domain_types["tableau"]}
    domain_parser.parse_effects(effects_ast=effects_tokens,
                                new_action=new_action,
                                domain_types=domain_types,
                                domain_functions=domain_functions,
                                domain_predicates=domain_predicates,
                                domain_constants=domain_consts)
    conditional_effect = new_action.conditional_effects.pop()
    assert conditional_effect is not None
    negative_condition = conditional_effect.negative_conditions.pop()
    add_effects = conditional_effect.add_effects
    assert negative_condition.untyped_representation == "(can-continue-group ?c ?to)"
    print([eff.untyped_representation for eff in add_effects])
    assert len(add_effects) == 2
    effects_str = {effect.untyped_representation for effect in add_effects}
    assert effects_str == {"(currently-updating-unmovable )", "(make-unmovable ?to)"}


def test_parse_effects_with_conditional_effects_with_two_conditions_and_two_effects_parse_correctly(
        domain_parser: DomainParser):
    conditional_effects = """(and (when
            (and (can-continue-group ?c ?to) (not (can-continue-group ?c ?from)))
            (and (currently-updating-unmovable) (make-unmovable ?to))))"""
    types_tokens = PDDLTokenizer(pddl_str=TEST_TYPES_FOR_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN).parse()
    constants_tokens = PDDLTokenizer(pddl_str=TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN).parse()
    effects_tokens = PDDLTokenizer(pddl_str=conditional_effects).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_consts = domain_parser.parse_constants(constants_tokens, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    new_action = Action()
    new_action.name = "test-action"
    new_action.signature = {"?c": domain_types["card"], "?from": domain_types["cardposition"],
                            "?fromdeal": domain_types["deal"], "?to": domain_types["card"],
                            "?totableau": domain_types["tableau"]}
    domain_parser.parse_effects(effects_ast=effects_tokens,
                                new_action=new_action,
                                domain_types=domain_types,
                                domain_functions=domain_functions,
                                domain_predicates=domain_predicates,
                                domain_constants=domain_consts)
    conditional_effect = new_action.conditional_effects.pop()
    assert conditional_effect is not None
    negative_condition = conditional_effect.negative_conditions.pop()
    positive_condition = conditional_effect.positive_conditions.pop()
    add_effects = conditional_effect.add_effects
    assert negative_condition.untyped_representation == "(can-continue-group ?c ?from)"
    assert positive_condition.untyped_representation == "(can-continue-group ?c ?to)"
    assert len(add_effects) == 2
    effects_str = {effect.untyped_representation for effect in add_effects}
    assert effects_str == {"(currently-updating-unmovable )", "(make-unmovable ?to)"}


def test_parse_action_with_conditional_effects_succeeds_in_parsing_action_and_returns_correct_fields(
        domain_parser: DomainParser):
    test_action_with_conditional_effects = """(deal-card
    :parameters (?c - card ?from - cardposition ?fromdeal - deal ?to - card ?totableau - tableau)
    :precondition
    (and
        (currently-dealing)
        (not (currently-updating-movable))
        (not (currently-updating-unmovable))
        (not (currently-updating-part-of-tableau))
        (not (currently-collecting-deck))
        (current-deal ?fromdeal)
        (TO-DEAL ?c ?totableau ?fromdeal ?from)
        (clear ?c)
        (on ?c ?from)
        (part-of-tableau ?to ?totableau)
        (clear ?to)
    )
    :effect
    (and
        (not (on ?c ?from))
        (on ?c ?to)
        (not (clear ?to))
        (clear ?from)
        (in-play ?c)
        (part-of-tableau ?c ?totableau)
        (movable ?c)
        (when
            (not (CAN-CONTINUE-GROUP ?c ?to))
            (and
                (currently-updating-unmovable)
                (make-unmovable ?to)
            )
        )
    )
    )"""
    types_tokens = PDDLTokenizer(pddl_str=TEST_TYPES_FOR_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN).parse()
    constants_tokens = PDDLTokenizer(pddl_str=TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN).parse()
    action_tokens = PDDLTokenizer(pddl_str=test_action_with_conditional_effects).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_consts = domain_parser.parse_constants(constants_tokens, domain_types)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    action = domain_parser.parse_action(action_tokens, domain_types, domain_functions, domain_predicates, domain_consts)
    assert action is not None
    assert action.name == "deal-card"
    assert action.conditional_effects is not None


def test_parse_conditional_effect_raises_error_if_conditional_effect_not_in_correct_format(
        domain_parser: DomainParser):
    conditional_effects_str = """
    (when (and (currently-updating-unmovable) (make-unmovable ?to)))"""
    types_tokens = PDDLTokenizer(pddl_str=TEST_TYPES_FOR_CONDITIONAL_DOMAIN).parse()
    constants_tokens = PDDLTokenizer(pddl_str=TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN).parse()
    conditional_effect_tokens = PDDLTokenizer(pddl_str=conditional_effects_str).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_consts = domain_parser.parse_constants(constants_tokens, domain_types)
    domain_functions = {}  # Functions are irrelevant for this case.
    action = Action()
    action.name = "deal-card"
    with raises(SyntaxError):
        domain_parser.parse_conditional_effect(conditional_effect_tokens, action, domain_functions, domain_consts)


def test_parse_action_with_universal_quantifier_in_conditional_effect_returns_correct_universal_quantifier_data(
        domain_parser: DomainParser):
    test_action_with_universal_quantifier = """
    (stop
      :parameters (?f - floor)
      :precondition (lift-at ?f)
      :effect (and 
                   (forall (?p - passenger) 
                      (when (and (boarded ?p) 
                                 (destin ?p ?f))
                            (and (not (boarded ?p)) 
                                 (served  ?p))))
    ))"""
    types_tokens = PDDLTokenizer(pddl_str=TYPES_FOR_UNIVERSAL_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_UNIVERSAL_QUANTIFIER_DOMAIN).parse()
    action_tokens = PDDLTokenizer(pddl_str=test_action_with_universal_quantifier).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}
    action = domain_parser.parse_action(action_ast=action_tokens, domain_types=domain_types,
                                        domain_functions=domain_functions, domain_predicates=domain_predicates)
    assert action is not None
    universal_effects = action.universal_effects
    assert len(universal_effects) == 1
    universal_effect = universal_effects.pop()
    assert universal_effect.quantified_parameter == "?p"
    assert universal_effect.quantified_type.name == "passenger"
    assert len(universal_effect.conditional_effect.add_effects) == 1
    assert len(universal_effect.conditional_effect.delete_effects) == 1
    assert len(universal_effect.conditional_effect.positive_conditions) == 2
    print(str(universal_effect))


def test_parse_action_with_two_universal_quantifiers_in_effect_extract_all_universal_effects(
        domain_parser: DomainParser):
    test_action_with_universal_quantifier = """
    (stop
      :parameters (?f - floor)
      :precondition (lift-at ?f)
      :effect (and 
               (forall (?p - passenger) 
                  (when (and (boarded ?p) 
                             (destin ?p ?f))
                        (and (not (boarded ?p)) 
                             (served  ?p))))
               (forall (?p - passenger)                
                   (when (and (origin ?p ?f) (not (served ?p)))
                              (boarded ?p)))))"""
    types_tokens = PDDLTokenizer(pddl_str=TYPES_FOR_UNIVERSAL_CONDITIONAL_DOMAIN).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=TEST_PREDICATES_FOR_UNIVERSAL_QUANTIFIER_DOMAIN).parse()
    action_tokens = PDDLTokenizer(pddl_str=test_action_with_universal_quantifier).parse()
    domain_types = domain_parser.parse_types(types_tokens)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = {}
    action = domain_parser.parse_action(action_ast=action_tokens, domain_types=domain_types,
                                        domain_functions=domain_functions, domain_predicates=domain_predicates)
    assert action is not None
    universal_effects = action.universal_effects
    assert len(universal_effects) == 2


def test_parse_simple_action_with_numeric_preconditions_and_effects_extracts_the_calculation_tree_correctly(
        domain_parser: DomainParser):
    test_simple_types = """(place locatable - object
	    depot distributor - place
        truck hoist surface - locatable
        pallet crate - surface)"""
    test_simple_predicates = """((at ?x - locatable ?y - place) 
             (on ?x - crate ?y - surface)
             (in ?x - crate ?y - truck)
             (lifting ?x - hoist ?y - crate)
             (available ?x - hoist)
             (clear ?x - surface))"""
    test_domain_simple_functions = """((load_limit ?t - truck) 
	(current_load ?t - truck) 
	(weight ?c - crate)
	(fuel-cost))
    """
    test_simple_numeric_action = """(Load
        :parameters (?x - hoist ?y - crate ?z - truck ?p - place)
        :precondition (and (at ?x ?p) (at ?z ?p) (lifting ?x ?y)
                (<= (+ (current_load ?z) (weight ?y)) (load_limit ?z)))
        :effect (and (not (lifting ?x ?y)) (in ?y ?z) (available ?x)
                (increase (current_load ?z) (weight ?x))))"""
    types_tokens = PDDLTokenizer(pddl_str=test_simple_types).parse()
    functions_tokens = PDDLTokenizer(pddl_str=test_domain_simple_functions).parse()
    predicate_tokens = PDDLTokenizer(pddl_str=test_simple_predicates).parse()
    action_tokens = PDDLTokenizer(pddl_str=test_simple_numeric_action).parse()

    domain_types = domain_parser.parse_types(types_tokens)
    domain_predicates = domain_parser.parse_predicates(predicate_tokens, domain_types)
    domain_functions = domain_parser.parse_functions(functions_tokens, domain_types)

    action = domain_parser.parse_action(action_tokens, domain_types, domain_functions, domain_predicates)
    assert len(action.numeric_preconditions) == 1
    precond_expression = action.numeric_preconditions.pop()
    assert precond_expression.root.id == "<="
    assert precond_expression.root.height == 2

    assert len(action.numeric_effects) == 1
    effect_expression = action.numeric_effects.pop()
    assert effect_expression.root.id == "increase"
    assert effect_expression.root.height == 1


def test_parse_simple_domain_with_only_boolean_actions_succeeds_in_parsing_all_domain_parts(
        domain_parser: DomainParser):
    domain = domain_parser.parse_domain()
    assert domain is not None
    assert len(domain.requirements) == 2
    assert len(domain.types) == 8
    assert len(domain.predicates) == 3
    assert len(domain.actions) == 6


def test_parse_woodworking_domain_with_boolean_actions_and_constants_succeeds_in_parsing_all_domain_parts():
    domain_parser = DomainParser(TEST_WOODWORKING_DOMAIN_PATH)
    domain = domain_parser.parse_domain()
    assert domain is not None
    assert len(domain.requirements) == 2
    assert len(domain.types) == 18
    assert len(domain.constants) == 11
    assert len(domain.predicates) == 14
    assert len(domain.actions) == 13


def test_parse_depot_domain_with_numeric_actions_succeeds_in_parsing_all_domain_parts():
    domain_parser = DomainParser(TEST_NUMERIC_DEPOT_DOMAIN_PATH)
    domain = domain_parser.parse_domain()
    assert domain is not None
    assert len(domain.requirements) == 2
    assert len(domain.types) == 10
    assert len(domain.predicates) == 6
    assert len(domain.functions) == 4
    assert len(domain.actions) == 5


def test_parse_action_with_function_equality_precondition_sets_a_new_numeric_precondition():
    domain_parser = DomainParser(PLANT_WATERING_DOMAIN)
    domain = domain_parser.parse_domain()
    assert domain is not None
    assert len(domain.actions) == 10
    assert len(domain.actions["pour"].numeric_preconditions) == 5
    for action in domain.actions.values():
        print(action.name)
        for numeric_precond in action.numeric_preconditions:
            print(str(numeric_precond))


def test_parse_preconditions_with_action_with_disjunctive_preconditions_extracts_different_set_of_preconditions():
    test_action_str = """(lift
	:parameters (?x - hoist ?y - crate ?z - surface ?p - place)
	:precondition (and (on ?y ?z) (at ?y ?p) (clear ?y) (at ?x ?p) (available ?x)
				(or (and		
    (>= (+ (* (fuel-cost ) -0.01) 0.78) 0.0)		
    (>= (+ (* (fuel-cost ) -0.02) 1.01) 0.0)		
    )		
    (and		
    (>= (+ (* (fuel-cost ) 0.01) -1.79) 0.0)		
    (>= (+ (+ (* (weight ?y) 0.01) (* (fuel-cost ) 0.01)) -2.05) 0.0)		
    (>= (+ (* (fuel-cost ) -0.01) 0.78) 0.0)		
    )))
        :effect (and (clear ?z) (lifting ?x ?y) (not (available ?x)) (not (on ?y ?z)) (not (at ?y ?p))
            (increase (fuel-cost ) 1.0)
    ))"""
    domain_parser = DomainParser(TEST_NUMERIC_DEPOT_DOMAIN_PATH)
    domain = domain_parser.parse_domain()
    action_tokens = PDDLTokenizer(pddl_str=test_action_str).parse()
    action = domain_parser.parse_action(action_tokens, domain.types, domain.functions, domain.predicates,
                                        domain.constants)
    disjunctive_preconditions = action.disjunctive_numeric_preconditions
    assert len(disjunctive_preconditions) == 2
    assert len(disjunctive_preconditions[0]) == 2
    assert len(disjunctive_preconditions[1]) == 3
    print(str(action))


def test_parse_domain_with_domain_with_conditional_effects_not_fail():
    domain_parser = DomainParser(SPIDER_DOMAIN_PATH)
    domain = domain_parser.parse_domain()
    print(str(domain))



"""Module that contains the parser for PDDL+ domain files."""
import logging
from pathlib import Path
from typing import List, Dict, Union, Tuple

from pddl_plus_parser.models import Domain, PDDLType, Predicate, PDDLConstant, PDDLFunction, Action, SignatureType, \
    NumericalExpressionTree, construct_expression_tree, ConditionalEffect, UniversalQuantifiedEffect
from .parsing_utils import parse_signature
from .pddl_tokenizer import PDDLTokenizer

ObjectType = PDDLType(name="object", parent=None)

COMPARISON_OPS = ["=", "<=", ">=", ">", "<"]
ASSIGNMENT_OPS = ["assign", "increase", "decrease"]
EQUALITY_OPERATOR = "="


class DomainParser:
    """Class that parses PDDL+ domain files."""

    tokenizer: PDDLTokenizer
    logger: logging.Logger
    partial_parsing: bool
    enable_disjunctions: bool

    def __init__(self, domain_path: Path, partial_parsing: bool = False, enable_disjunctions: bool = False):
        self.tokenizer = PDDLTokenizer(domain_path)
        self.logger = logging.getLogger(__name__)
        self.partial_parsing = partial_parsing
        self.enable_disjunctions = enable_disjunctions

    def parse_types(self, types: List[str]) -> Dict[str, PDDLType]:
        """Parse the types of the domain.

        :param types: the list containing the
        :return: a mapping between the type name and the appropriate PDDLType object.
        """
        self.logger.info("Starting to parse the types in the domain!")
        pddl_types = {}
        same_types_objects = []
        index = 0
        while index < len(types):
            if types[index] != "-":
                same_types_objects.append(types[index])
                index += 1
                continue

            pddl_type = types[index + 1]
            parent_type = pddl_types.get(pddl_type, PDDLType(name=pddl_type, parent=ObjectType))
            pddl_types.update({descendant_typ_name: PDDLType(name=descendant_typ_name, parent=parent_type)
                               for descendant_typ_name in same_types_objects})
            same_types_objects = []
            index += 2
            continue

        if len(same_types_objects) > 0:
            pddl_types.update({type_name: PDDLType(name=type_name, parent=ObjectType)
                               for type_name in same_types_objects})

        pddl_types["object"] = ObjectType
        self.logger.debug(f"Extracted {len(pddl_types)} types while parsing the types AST.")
        return pddl_types

    def parse_constants(self, constants_ast: List[str], domain_types: Dict[str, PDDLType]) -> Dict[str, PDDLConstant]:
        """Parses the constants that appear in the constants part of the domain AST.

        :param constants_ast: the constants that were parsed in the domain AST.
        :param domain_types: the types that exist in the domain.
        :return: a list containing all typed constants.
        """
        self.logger.info("Starting to parse the constants of the domain!")
        constants = {}
        same_type_constants = []
        type_marker_reached = False
        for constant_name in constants_ast:
            if type_marker_reached:
                # This means that the name of the constant here is actually its type.
                if constant_name not in domain_types:
                    raise SyntaxError("Received invalid type for the constants!")

                constants.update(
                    {name: PDDLConstant(name, domain_types[constant_name]) for name in same_type_constants})
                type_marker_reached = False
                same_type_constants = []
                continue

            if constant_name == "-":
                type_marker_reached = True
                continue

            same_type_constants.append(constant_name)

        self.logger.debug(f"Extracted {len(constants)} from the domain.")
        return constants

    def parse_predicate(self, predicate: List[str], domain_types: Dict[str, PDDLType]) -> Predicate:
        """Parse a single predicate from an AST representation in the PDDL domain.

        :param predicates_ast: the predicate in the form of an AST list of strings.
        :param domain_types: the types that were extracted from the domain.
        :return: the predicate object that represents the list of strings that were given.
        """
        self.logger.info(f"Parsing the predicate represented by the AST - {predicate}")
        predicate_name = predicate[0]
        if (len(predicate[1:]) % 3) != 0:
            raise SyntaxError(f"Received a predicate with a wrong signature - {predicate[1:]}")

        signature_items = iter(predicate[1:])
        predicate_signature = parse_signature(signature_items, domain_types)
        extracted_predicate = Predicate(name=predicate_name, signature=predicate_signature)
        self.logger.debug(f"Finished extracting the predicate - {extracted_predicate}")
        return extracted_predicate

    def parse_untyped_predicate(self, untyped_predicate: List[str], action_signature: SignatureType,
                                domain_constants: Dict[str, PDDLConstant] = {}) -> Predicate:
        """Parse an untyped predicate that appears in actions.

        :param untyped_predicate: the untyped predicate that needs to be matched to the typed predicate from the
            predicates definitions.
        :param action_signature: the signed signature of the action.
        :return: the predicate including the relevant types.
        :param domain_constants: the constants that are defined in the domain.
        """
        self.logger.info(f"Parsing the untyped action predicate represented by the AST - {untyped_predicate}")
        predicate_name = untyped_predicate[0]
        possible_signed_objects = {key: val for key, val in action_signature.items()}
        possible_signed_objects.update({const_name: const.type for const_name, const in domain_constants.items()})
        # Since we assume that the order in maintained in the predicates we can match the signatures.
        signed_signature = {parameter_name: possible_signed_objects[parameter_name] for
                            parameter_name in untyped_predicate[1:]}

        signed_action_predicate = Predicate(name=predicate_name, signature=signed_signature)
        self.logger.debug(f"Extracted the predicate - {signed_action_predicate}")
        return signed_action_predicate

    def parse_predicates(self, predicates_ast: List[List[str]],
                         domain_types: Dict[str, PDDLType]) -> Dict[str, Predicate]:
        """Parses the predicates that appear in the predicates parsed AST.

        :param predicates_ast: the AST that contains the predicates of the domain.
        :param domain_types: the types that exist in the domain.
        :return: a mapping between the predicate name and the predicate itself.
        """
        predicates = {}
        for predicate in predicates_ast:
            if predicate[0] == ":private":
                for private_predicate in predicate[1:]:
                    extracted_private_predicate = self.parse_predicate(private_predicate, domain_types)
                    predicates[extracted_private_predicate.name] = extracted_private_predicate

                continue

            extracted_predicate = self.parse_predicate(predicate, domain_types)
            predicates[extracted_predicate.name] = extracted_predicate

        return predicates

    def parse_functions(self, functions_ast: List[List[str]],
                        domain_types: Dict[str, PDDLType]) -> Dict[str, PDDLFunction]:
        """Parses the functions that appear in the functions parsed AST.

        :param functions_ast: the AST that contains the numerical functions of the domain.
        :param domain_types: the types that exist in the domain.
        :return: a mapping between a function name and the function itself.
        """
        self.logger.info("Starting to parse the function' definition in the domain data.")
        functions = {}
        for function_items in functions_ast:
            function_name = function_items[0]
            if (len(function_items[1:]) % 3) != 0:
                raise SyntaxError(f"Received a function with a wrong signature - {function_items[1:]}")

            signature_items = iter(function_items[1:])
            function_signature = parse_signature(signature_items, domain_types)
            functions[function_name] = PDDLFunction(name=function_name, signature=function_signature)

        return functions

    def parse_disjunctive_numeric_preconditions(
            self, numeric_preconditions_node: List[Union[str, List[str]]],
            new_action: Action, domain_functions: Dict[str, PDDLFunction]) -> None:
        """Parse a set of disjunctive numeric preconditions.

        :param numeric_preconditions_node: the node that contains the numeric preconditions.
        :param new_action: the action that is currently being parsed.
        :param domain_functions: the functions that are defined in the domain.
        """
        self.logger.info("Starting to parse the disjunctive numeric preconditions!")
        for conditions_set_ast in numeric_preconditions_node:
            if conditions_set_ast[0] != "and":
                raise SyntaxError(
                    f"Only accepting conjunctive preconditions! Action - {new_action.name} does not conform!")

            conditions_set = {NumericalExpressionTree(construct_expression_tree(node, domain_functions)) for node in
                              conditions_set_ast[1:]}
            new_action.disjunctive_numeric_preconditions.append(conditions_set)

    def parse_preconditions(self, preconditions_ast: List[Union[str, List[str]]], new_action: Action,
                            domain_functions: Dict[str, PDDLFunction],
                            domain_predicates: Dict[str, Predicate],
                            domain_constants: Dict[str, PDDLConstant]) -> None:
        """Parse the preconditions of a single action.

        :param preconditions_ast: the AST representation of the action's preconditions.
        :param new_action: the action that is currently being parsed.
        :param domain_functions: the functions that exist in the domain.
        :param domain_predicates: the predicates that exist in the domain.
        :param domain_constants: the constants that might exist in the domain.
        """
        new_action.positive_preconditions = set()
        new_action.negative_preconditions = set()
        new_action.equality_preconditions = set()
        new_action.inequality_preconditions = set()
        new_action.numeric_preconditions = set()
        if len(preconditions_ast) == 0:
            self.logger.warning("Received an action with no preconditions.")
            return

        if preconditions_ast[0] != "and" and not self.enable_disjunctions and len(preconditions_ast[1:]) > 1:
            raise SyntaxError(f"Only accepting conjunctive preconditions! Action - {new_action.name} does not conform!")

        for precondition_node in preconditions_ast[1:]:
            if precondition_node[0] in domain_predicates:
                new_action.positive_preconditions.add(
                    self.parse_untyped_predicate(precondition_node, new_action.signature, domain_constants))
                continue

            if precondition_node[0] == "not":
                inner_node = precondition_node[1]
                if inner_node[0] == EQUALITY_OPERATOR:
                    self.logger.debug("Adding new lifted objects that should be tested for inequality")
                    new_action.inequality_preconditions.add((inner_node[1], inner_node[2]))
                    continue

                new_action.negative_preconditions.add(
                    self.parse_untyped_predicate(precondition_node[1], new_action.signature, domain_constants))
                continue

            if precondition_node[0] == EQUALITY_OPERATOR:
                if not isinstance(precondition_node[1], List):
                    self.logger.debug("Adding new lifted objects that should be tested for equality")
                    new_action.equality_preconditions.add((precondition_node[1], precondition_node[2]))
                    continue

            if precondition_node[0] == "or":
                self.logger.debug("Assuming that the OR operator is used for the disjunction of numeric preconditions.")
                self.parse_disjunctive_numeric_preconditions(precondition_node[1:], new_action, domain_functions)

            if precondition_node[0] in COMPARISON_OPS:
                numerical_precondition = NumericalExpressionTree(
                    construct_expression_tree(precondition_node, domain_functions))
                new_action.numeric_preconditions.add(numerical_precondition)
                continue

    def _parse_single_conditional(self, conditional: List[Union[str, List[str]]],
                                  positive_conditionals: List[Predicate], negative_conditionals: List[Predicate],
                                  numeric_conditionals: List[NumericalExpressionTree],
                                  signature: Dict[str, PDDLType], domain_functions: Dict[str, PDDLFunction],
                                  domain_constants: Dict[str, PDDLConstant]) -> None:
        """Parse a single conditional from the conditional effect.

        :param conditional: the single conditional effect to parse.
        :param positive_conditionals: the positive conditionals of the conditional effect to
            possibly add the conditional to.
        :param negative_conditionals: the negative conditionals of the conditional effect to
            possibly add the conditional to.
        :param numeric_conditionals: the numeric conditionals of the conditional effect.
        :param signature: the action's signature.
        :param domain_functions: the functions that exist in the domain.
        :param domain_constants: the constants that exist in the domain.
        """
        if conditional[0] == "not":
            negative_conditionals.append(
                self.parse_untyped_predicate(conditional[1], signature, domain_constants))

        elif conditional[0] in COMPARISON_OPS:
            numerical_precondition = NumericalExpressionTree(
                construct_expression_tree(conditional, domain_functions))
            numeric_conditionals.append(numerical_precondition)

        else:
            positive_conditionals.append(self.parse_untyped_predicate(conditional, signature, domain_constants))

    def _parse_antecedents(self, antecedent_ast: List[Union[str, List[str]]],
                           signature: Dict[str, PDDLType], domain_functions: Dict[str, PDDLFunction],
                           domain_constants: Dict[str, PDDLConstant]) -> Tuple[
        List[Predicate], List[Predicate], List[NumericalExpressionTree]]:
        """Parse the antecedents for a conditional effect of an action.

        :param antecedent_ast: the AST representation of the conditionals.
        :param signature: the action's signature.
        :param domain_functions: the functions that exist in the domain.
        :param domain_constants: the constants that exist in the domain.
        :return: a tuple containing the positive and negative and numeric conditionals (in this order).
        """
        positive_conditionals = []
        negative_conditionals = []
        numeric_conditionals = []
        if antecedent_ast[0] == "and":
            self.logger.debug("Received a conjunctive conditional effect with more than one condition.")
            for conditional in antecedent_ast[1:]:
                self._parse_single_conditional(conditional, positive_conditionals, negative_conditionals,
                                               numeric_conditionals, signature, domain_functions, domain_constants)

        else:
            self._parse_single_conditional(antecedent_ast, positive_conditionals, negative_conditionals,
                                           numeric_conditionals, signature, domain_functions, domain_constants)

        return positive_conditionals, negative_conditionals, numeric_conditionals

    def _parse_single_conditional_effect(self, action, conditional_effect_ast, add_effects, del_effects,
                                         numeric_effects, domain_constants, domain_functions) -> None:
        """

        :param action: the action that is being parsed.
        :param conditional_effect_ast: the AST representation of the conditional effect.
        :param add_effects: the add effects of the action.
        :param del_effects: the delete effects of the action.
        :param numeric_effects: the numeric effects of the action.
        :param domain_constants: the constants that exist in the domain.
        :param domain_functions: the functions that exist in the domain.
        """

        if conditional_effect_ast[0] == "not":
            del_effects.append(
                self.parse_untyped_predicate(conditional_effect_ast[1], action.signature, domain_constants))

        elif conditional_effect_ast[0] in ASSIGNMENT_OPS:
            numeric_effects.append(
                NumericalExpressionTree(construct_expression_tree(conditional_effect_ast, domain_functions)))

        else:
            add_effects.append(self.parse_untyped_predicate(conditional_effect_ast, action.signature, domain_constants))

    def _construct_conditional_effects(
            self, conditional_effect_ast: List[Union[str, List[str]]],
            positive_conditionals: List[Predicate],
            negative_conditionals: List[Predicate],
            numeric_conditionals: List[NumericalExpressionTree],
            action: Action,
            domain_functions: Dict[str, PDDLFunction],
            domain_constants: Dict[str, PDDLConstant]) -> ConditionalEffect:
        """Parse all the conditional effects that are under the same when statement - can be composite.

        :param conditional_effect_ast: the AST representation of the conditional effects.
        :param positive_conditionals: the positive conditionals of the conditional effect.
        :param negative_conditionals: the negative conditionals of the conditional effect.
        :param numeric_conditionals: the numeric conditionals of the conditional effect.
        :param action: the action that is being parsed.
        :param domain_functions: the functions that exist in the domain.
        :param domain_constants: the constants that exist in the domain.
        """
        add_effects = []
        del_effects = []
        numeric_effects = []
        if conditional_effect_ast[0] == "and":
            for effect in conditional_effect_ast[1:]:
                self._parse_single_conditional_effect(action, effect, add_effects, del_effects, numeric_effects,
                                                      domain_constants, domain_functions)

        else:
            self._parse_single_conditional_effect(action, conditional_effect_ast, add_effects, del_effects,
                                                  numeric_effects, domain_constants, domain_functions)

        conditional_effect = ConditionalEffect()
        conditional_effect.add_effects = set(add_effects)
        conditional_effect.delete_effects = set(del_effects)
        conditional_effect.numeric_effects = set(numeric_effects)
        conditional_effect.positive_conditions = set(positive_conditionals)
        conditional_effect.negative_conditions = set(negative_conditionals)
        conditional_effect.numeric_conditions = set(numeric_conditionals)
        return conditional_effect

    def _extend_action_to_contain_quantified_objects(
            self, action: Action, quantified_parameter_name: str, quantified_type: PDDLType) -> Action:
        """Extends the relevant parts of the action to include the quantified object.

        :param action: the action to extend.
        :param quantified_parameter_name: the name of the parameter that is added to the extended action.
        :param quantified_type: the type of the quatified object.
        :return: the extended action.
        """
        combined_signature = {**action.signature, quantified_parameter_name: quantified_type}
        extended_action = Action()
        extended_action.name = action.name
        extended_action.signature = combined_signature
        return extended_action

    def parse_conditional_effect(
            self, conditional_effect_ast: List[Union[str, List[str]]], action: Action,
            domain_functions: Dict[str, PDDLFunction],
            domain_constants: Dict[str, PDDLConstant]) -> None:
        """Parse a conditional effect of an action.

        :param conditional_effect_ast: the AST representation of the conditional effect.
        :param action: the action that is being parsed.
        :param domain_functions: the functions that exist in the domain.
        :param domain_constants: the constants that exist in the domain.
        """
        self.logger.debug("Parsing conditional effect node.")
        if len(conditional_effect_ast[1:]) != 2:
            raise SyntaxError(f"Conditional effect scheme does not match for action {action.name}!")

        positive_conditionals, negative_conditionals, numeric_conditionals = self._parse_antecedents(
            conditional_effect_ast[1], action.signature, domain_functions, domain_constants)
        conditional_effect = self._construct_conditional_effects(conditional_effect_ast[2], positive_conditionals,
                                                                 negative_conditionals,
                                                                 numeric_conditionals, action, domain_functions,
                                                                 domain_constants)
        action.conditional_effects.add(conditional_effect)

    def parse_universally_quantified_effect(
            self, universal_quantifier_ast: List[Union[str, List[str]]],
            action: Action,
            domain_types: Dict[str, PDDLType],
            domain_functions: Dict[str, PDDLFunction],
            domain_constants: Dict[str, PDDLConstant]) -> None:
        """Parse a universally quantified effect of an action.

        :param universal_quantifier_ast: the ast representation of the universally quantified effect.
        :param action: the action that is being parsed.
        :param domain_types: the types that exist in the domain.
        :param domain_functions: the functions that exist in the domain.
        :param domain_constants: the constants that exist in the domain.
        """
        self.logger.debug("Parsing universally quantified effect node.")
        if len(universal_quantifier_ast) != 2:
            raise SyntaxError(f"Universal quantifier scheme does not match for action {action.name}!")

        quantified_object_components = universal_quantifier_ast[0]
        # quantified object components are of the form ( ?x - type )
        if len(quantified_object_components) != 3:
            raise SyntaxError(f"Quantified object scheme does not match for action {action.name}!"
                              f"Expected 3 components, got {len(quantified_object_components)}!")

        quantified_parameter_name = quantified_object_components[0]
        quantified_type = domain_types[quantified_object_components[2]]
        extended_action = self._extend_action_to_contain_quantified_objects(
            action, quantified_parameter_name, quantified_type)

        conditional_effect_ast = universal_quantifier_ast[1]
        positive_conditionals, negative_conditionals, numeric_conditionals = self._parse_antecedents(
            conditional_effect_ast[1], extended_action.signature, domain_functions, domain_constants)
        conditional_effect = self._construct_conditional_effects(conditional_effect_ast[2],
                                                                 positive_conditionals,
                                                                 negative_conditionals,
                                                                 numeric_conditionals,
                                                                 extended_action,
                                                                 domain_functions,
                                                                 domain_constants)

        universal_quantifier = UniversalQuantifiedEffect(quantified_parameter=quantified_parameter_name,
                                                         quantified_type=quantified_type,
                                                         conditional_effect=conditional_effect)
        action.universal_effects.add(universal_quantifier)

    def parse_effects(self, effects_ast: List[Union[str, List[str]]],
                      new_action: Action,
                      domain_types: Dict[str, PDDLType],
                      domain_functions: Dict[str, PDDLFunction],
                      domain_predicates: Dict[str, Predicate],
                      domain_constants: Dict[str, PDDLConstant]) -> None:
        """Parse the effects of a single action.

        :param effects_ast: the AST representation of the action's effects.
        :param new_action: the action that is currently being parsed.
        :param domain_types: the types that exist in the domain.
        :param domain_functions: the functions that exist in the domain.
        :param domain_predicates: the predicates that are defined in the domain.
        :param domain_constants: the domains that might exist in the domain.
        """
        new_action.add_effects = set()
        new_action.delete_effects = set()
        new_action.numeric_effects = set()
        if effects_ast[0] != "and":
            print(effects_ast[0])
            raise SyntaxError(
                f"Only accepting conjunctive effects! Action - {new_action.name} does not conform!")

        for effect_node in effects_ast[1:]:
            if effect_node[0] in domain_predicates:
                new_action.add_effects.add(
                    self.parse_untyped_predicate(effect_node, new_action.signature, domain_constants))
                continue

            if effect_node[0] == "not":
                new_action.delete_effects.add(
                    self.parse_untyped_predicate(effect_node[1], new_action.signature, domain_constants))
                continue

            if effect_node[0] == "forall":
                self.parse_universally_quantified_effect(effect_node[1:], new_action, domain_types, domain_functions,
                                                         domain_constants)
                continue

            if effect_node[0] == "when":
                self.parse_conditional_effect(effect_node, new_action, domain_functions, domain_constants)
                continue

            if effect_node[0] in ASSIGNMENT_OPS:
                numerical_precondition = NumericalExpressionTree(
                    construct_expression_tree(effect_node, domain_functions))
                new_action.numeric_effects.add(numerical_precondition)
                continue

    def parse_action(self, action_ast: List[Union[str, List[str]]], domain_types: Dict[str, PDDLType],
                     domain_functions: Dict[str, PDDLFunction],
                     domain_predicates: Dict[str, Predicate],
                     domain_constants: Dict[str, PDDLConstant] = {}) -> Action:
        """Parse a single action AST and returns the object that represent a single action.

        :param action_ast: the AST representation of the action.
        :param domain_types: the types that were extracted from the domain.
        :param domain_functions: the functions that were extracted from the AST.
        :param domain_predicates: the predicates that were extracted from the domain's AST.
        :param domain_constants: the constants that might exist in the domain and might appear in the
            preconditions / effects.
        :return: the action object that is generated using the data from the domain AST.
        """
        self.logger.info("Starting to parse a new action!")
        new_action = Action()
        new_action.name = action_ast[0].lower()
        if len(action_ast[1:]) != 6:  # the number of different sections for the action.
            raise SyntaxError(f"Received an Illegal action AST definition! The action given - {action_ast}")

        action_section_iterator = iter(action_ast[1:])
        for action_label_item in action_section_iterator:
            if action_label_item == ":parameters":
                self.logger.debug(f"Parsing the parameters of the action - {new_action.name}")
                parameters_list = next(action_section_iterator)
                new_action.signature = parse_signature(iter(parameters_list), domain_types)
                continue

            if action_label_item == ":precondition" and not self.partial_parsing:
                self.logger.debug(f"Starting to parse the preconditions of the action - {new_action.name}")
                self.parse_preconditions(next(action_section_iterator), new_action, domain_functions,
                                         domain_predicates, domain_constants)
                continue

            if action_label_item == ":effect" and not self.partial_parsing:
                self.logger.debug(f"Starting to parse the effects of the action - {new_action.name}")
                self.parse_effects(next(action_section_iterator), new_action, domain_types, domain_functions,
                                   domain_predicates, domain_constants)

        return new_action

    def parse_domain(self) -> Domain:
        """The main entry point that parses the domain file and returns the resulting Domain object.

        :return: the domain object extracted from the tokens.
        """
        domain_expressions = self.tokenizer.parse()
        if domain_expressions[0] != "define":
            raise SyntaxError("Encountered a PDDL that does not start with 'define' statement!")

        domain = Domain()
        for expression in domain_expressions[1:]:
            if expression[0] == "domain":
                domain.name = expression[1]

            elif expression[0] == ":requirements":
                domain.requirements = expression[1:]

            elif expression[0] == ":types":
                domain.types = self.parse_types(expression[1:])

            elif expression[0] == ":constants":
                domain.constants = self.parse_constants(expression[1:], domain.types)

            elif expression[0] == ":predicates":
                domain.predicates = self.parse_predicates(expression[1:], domain.types)

            elif expression[0] == ":functions":
                domain.functions = self.parse_functions(expression[1:], domain.types)

            elif expression[0] == ":action":
                new_action: Action = self.parse_action(expression[1:], domain.types, domain.functions,
                                                       domain.predicates, domain.constants)
                domain.actions[new_action.name] = new_action

            elif expression[0] == ":process" or expression[0] == ":event":
                self.logger.debug("Still no support for temporal domains.")
                # TODO: complete once I finish numeric actions support

        return domain





"""module to represent an operator that can apply actions and change state objects."""
import logging
from collections import defaultdict
from typing import List, Set, Dict, Tuple, Optional

from anytree import AnyNode

from . import PDDLObject
from .conditional_effect import ConditionalEffect
from .numerical_expression import NumericalExpressionTree, evaluate_expression
from .pddl_action import Action
from .pddl_domain import Domain
from .pddl_function import PDDLFunction
from .pddl_predicate import GroundedPredicate, Predicate, SignatureType
from .pddl_state import State
from .universal_quantifier import UniversalQuantifiedEffect


def set_expression_value(expression_node: AnyNode, state_fluents: Dict[str, PDDLFunction]) -> None:
    """Set the value of the expression according to the fluents present in the state.

    :param expression_node: the node that is currently being observed.
    :param state_fluents: the grounded numeric fluents present in the state.
    """
    if expression_node.is_leaf:
        if not isinstance(expression_node.value, PDDLFunction):
            return

        grounded_fluent: PDDLFunction = expression_node.value
        try:
            grounded_fluent.set_value(state_fluents[grounded_fluent.untyped_representation].value)

        except KeyError:
            grounded_fluent.set_value(0.0)

        return

    set_expression_value(expression_node.children[0], state_fluents)
    set_expression_value(expression_node.children[1], state_fluents)


class Operator:
    action: Action
    domain: Domain
    grounded_call_objects: List[str]
    grounded: bool
    logger: logging.Logger

    # These fields are constructed after the grounding process.
    grounded_positive_preconditions: Set[GroundedPredicate]
    grounded_negative_preconditions: Set[GroundedPredicate]
    grounded_equality_preconditions: Set[Tuple[str, str]]
    grounded_inequality_preconditions: Set[Tuple[str, str]]
    grounded_numeric_preconditions: Set[NumericalExpressionTree]
    grounded_add_effects: Set[GroundedPredicate]
    grounded_delete_effects: Set[GroundedPredicate]
    grounded_numeric_effects: Set[NumericalExpressionTree]
    grounded_conditional_effects: Set[ConditionalEffect]
    lifted_universal_effects: Set[UniversalQuantifiedEffect]
    grounded_disjunctive_numeric_preconditions: List[Set[NumericalExpressionTree]]
    problem_objects: Dict[str, PDDLObject]

    def __init__(self, action: Action, domain: Domain, grounded_action_call: List[str],
                 problem_objects: Optional[Dict[str, PDDLObject]] = None):
        self.action = action
        self.domain = domain
        self.grounded_call_objects = grounded_action_call
        self.grounded = False
        self.problem_objects = problem_objects
        self.logger = logging.getLogger(__name__)

    @property
    def name(self) -> str:
        return self.action.name

    @property
    def typed_action_call(self) -> str:
        signature_str_items = [f"{parameter_name} - {str(parameter_type)}"
                               for parameter_name, parameter_type in
                               zip(self.grounded_call_objects, self.action.signature.values())]
        return f"({self.name} {' '.join(signature_str_items)})"

    def __str__(self):
        called_objects = " ".join(self.grounded_call_objects)
        return f"({self.name} {called_objects})"

    def _ground_predicates(self, lifted_predicates: Set[Predicate],
                           parameters_map: Dict[str, str]) -> Set[GroundedPredicate]:
        """Grounds predicates that appear in the grounded operator.

        :param lifted_predicates: the lifted predicates definition originating from the domain.
        :param parameters_map: the mapping between the action's parameters and the objects that the action
            was actually called with.
        :return: the predicates that appear in the action containing concrete objects.
        """
        output_grounded_predicates = set()
        for predicate in lifted_predicates:
            predicate_name = predicate.name
            # I want the grounded predicate to have the same signature as the original signature so that I can later
            # on efficiently search for it in the states.
            predicate_signature = {param: param_type for param, param_type in
                                   self.domain.predicates[predicate_name].signature.items()}
            predicate_params = list(predicate.signature.keys())
            if len(self.domain.constants) > 0:
                predicate_params.extend(list(self.domain.constants.keys()))

            lifted_predicate_params = [param for param in predicate.signature]
            predicate_object_mapping = {}
            for index, parameter_name in enumerate(predicate_signature):
                if predicate_params[index] in self.domain.constants:
                    predicate_object_mapping[parameter_name] = predicate_params[index]

                else:
                    predicate_object_mapping[parameter_name] = parameters_map[predicate_params[index]]

            # Matching the types to be the same as the ones in the action.
            self._fix_grounded_predicate_types(lifted_predicate_params, predicate_signature)

            output_grounded_predicates.add(GroundedPredicate(name=predicate_name,
                                                             signature=predicate_signature,
                                                             object_mapping=predicate_object_mapping))
        return output_grounded_predicates

    def _fix_grounded_predicate_types(self, lifted_predicate_params: List[str],
                                      predicate_signature: SignatureType) -> None:
        """Fix the types of the grounded predicate to match those in the action itself.

        :param lifted_predicate_params: the names of the lifted predicate parameters.
        :param predicate_signature: the signature of the grounded predicate.
        """
        for domain_def_parameter, lifted_predicate_param_name in zip(predicate_signature, lifted_predicate_params):
            if lifted_predicate_param_name in self.domain.constants:
                predicate_signature[domain_def_parameter] = self.domain.constants[lifted_predicate_param_name].type

            else:
                predicate_signature[domain_def_parameter] = self.action.signature[lifted_predicate_param_name]

    def _iterate_calc_tree_and_ground(self, calc_node: AnyNode, parameters_map: Dict[str, str]) -> AnyNode:
        """Recursion function that iterates over the lifted calculation tree and grounds its elements.

        :param calc_node: the current node the recursion currently visits.
        :param parameters_map: the mapping between the action parameters and the objects in which the action was called.
        :return: the node that represents the calculations of the current lifted expression.
        """
        if calc_node.is_leaf:
            if isinstance(calc_node.value, PDDLFunction):
                lifted_function: PDDLFunction = calc_node.value
                lifted_function_params = [param for param in lifted_function.signature]
                grounded_signature = {}
                for index, parameter_name in enumerate(lifted_function_params):
                    if parameter_name in self.domain.constants:
                        grounded_signature[parameter_name] = lifted_function.signature[parameter_name]

                    else:
                        grounded_signature[parameters_map[lifted_function_params[index]]] = \
                            lifted_function.signature[parameter_name]

                grounded_function = PDDLFunction(name=lifted_function.name,
                                                 signature=grounded_signature)
                return AnyNode(id=str(grounded_function), value=grounded_function)

            return AnyNode(id=calc_node.id, value=calc_node.value)

        return AnyNode(
            id=calc_node.id, value=calc_node.value, children=[
                self._iterate_calc_tree_and_ground(calc_node.children[0], parameters_map),
                self._iterate_calc_tree_and_ground(calc_node.children[1], parameters_map),
            ])

    def _ground_numeric_calculation_tree(self, lifted_numeric_exp_tree: NumericalExpressionTree,
                                         parameters_map: Dict[str, str]) -> NumericalExpressionTree:
        """grounds a calculation expression and returns the version containing the objects instead of the parameters.

        :param lifted_numeric_exp_tree: the lifted calculation tree.
        :param parameters_map: the mapping between the action's parameters and the objects in which the action
            was called with.
        :return: the grounded expression tree.
        """
        root = lifted_numeric_exp_tree.root
        grounded_root = self._iterate_calc_tree_and_ground(root, parameters_map)
        return NumericalExpressionTree(expression_tree=grounded_root)

    def _ground_numeric_expressions(self, lifted_numeric_exp_tree: Set[NumericalExpressionTree],
                                    parameters_map: Dict[str, str]) -> Set[NumericalExpressionTree]:
        """Grounds a set of numeric expressions.

        :param lifted_numeric_exp_tree: the set containing the numeric expressions to ground.
        :param parameters_map: the mapping between the action's parameters and the objects using which the action was
            called.
        :return: a set containing the grounded expressions.
        """
        grounded_numeric_expressions = set()
        for expression in lifted_numeric_exp_tree:
            grounded_numeric_expressions.add(self._ground_numeric_calculation_tree(expression, parameters_map))

        return grounded_numeric_expressions

    def _ground_conditional_effect(self, lifted_condition: ConditionalEffect,
                                   parameters_map: Dict[str, str]) -> ConditionalEffect:
        """Grounds a single conditional effect.

        :param lifted_condition: the conditional effect to ground.
        :param parameters_map: the mapping between the action's parameters and the objects using which the action was
            called.
        :return: a grounded conditional effect.
        """
        grounded_conditional_effect = ConditionalEffect()
        grounded_conditional_effect.positive_conditions = self._ground_predicates(
            lifted_condition.positive_conditions, parameters_map)
        grounded_conditional_effect.negative_conditions = self._ground_predicates(lifted_condition.negative_conditions,
                                                                                  parameters_map)
        grounded_conditional_effect.numeric_conditions = self._ground_numeric_expressions(
            lifted_condition.numeric_conditions, parameters_map)
        grounded_conditional_effect.add_effects = self._ground_predicates(lifted_condition.add_effects, parameters_map)
        grounded_conditional_effect.delete_effects = self._ground_predicates(lifted_condition.delete_effects,
                                                                             parameters_map)
        grounded_conditional_effect.numeric_effects = self._ground_numeric_expressions(lifted_condition.numeric_effects,
                                                                                       parameters_map)
        return grounded_conditional_effect

    @staticmethod
    def _ground_equality_objects(equality_preconditions: Set[Tuple[str, str]],
                                 parameters_map: Dict[str, str]) -> Set[Tuple[str, str]]:
        """Grounds the in/equality operators from the preconditions.

        :param equality_preconditions: the set of lifted signature items that are to be tested for equality.
        :param parameters_map: the mapping between the lifted and the grounded objects.
        :return: the grounded objects that should/n't be equal.
        """
        return {(parameters_map[obj1], parameters_map[obj2]) for obj1, obj2 in equality_preconditions}

    def _positive_preconditions_hold(self, state: State, conditions: Optional[Set[GroundedPredicate]] = None) -> bool:
        """Check whether the positive preconditions hold in the given state.

        :param state: the state which the action is applied on.
        :param conditions: the positive preconditions to check.
        :return: whether the positive preconditions hold.
        """
        positive_conditions = conditions or self.grounded_positive_preconditions
        self.logger.info(
            f"Validating whether or not the positive state variables matching {self.__str__()} grounded predicates.")
        for positive_precondition in positive_conditions:
            try:
                state_grounded_predicates = state.state_predicates[positive_precondition.lifted_untyped_representation]
                untyped_predicates = [p.untyped_representation for p in state_grounded_predicates]
                if positive_precondition.untyped_representation not in untyped_predicates:
                    self.logger.debug(f"Did not find the grounded predicate "
                                      f"{positive_precondition.untyped_representation}")
                    return False

            except KeyError:
                self.logger.debug(f"Did not find the predicate {positive_precondition.lifted_untyped_representation}")
                return False

        self.logger.debug("All positive preconditions were found in the state.")
        return True

    def _negative_preconditions_hold(self, state: State, conditions: Optional[Set[GroundedPredicate]] = None) -> bool:
        """Check whether the negative preconditions hold in the given state.

        :param state: the state which the action is applied on.
        :param conditions: the negative preconditions to check.
        :return: whether the negative preconditions hold.
        """
        negative_conditions = conditions or self.grounded_negative_preconditions
        self.logger.info("Validating that all of the negative preconditions don't exist in the state.")
        for negative_precondition in negative_conditions:
            try:
                state_grounded_predicates = state.state_predicates[negative_precondition.lifted_untyped_representation]
                untyped_predicates = [p.untyped_representation for p in state_grounded_predicates]
                if negative_precondition.untyped_representation in untyped_predicates:
                    self.logger.debug(
                        f"Found the predicate {str(negative_precondition)} but it should not hold in the state!")
                    return False

            except KeyError:
                continue

        self.logger.debug("All negative preconditions do not hold in the state.")
        return True

    @staticmethod
    def _equality_holds(grounded_objects: Set[Tuple[str, str]]) -> bool:
        """validates if all the requested objects are equal.

        :param grounded_objects: the objects that are being tested.
        :return: whether they are equal.
        """
        return all([obj[0] == obj[1] for obj in grounded_objects])

    def _numeric_conditions_set_hold(self, state: State, numeric_conditions: Set[NumericalExpressionTree]) -> bool:
        """Check if the set of numeric conditions holds as a whole.

        :param state: the state that the action is being applied to.
        :param numeric_conditions: the set of numeric conditions to check.
        :return: whether the set of numeric conditions holds.
        """
        for grounded_expression in numeric_conditions:
            try:
                self.logger.debug("Setting the values of the numeric state functions to the grounded operator")
                set_expression_value(grounded_expression.root, state.state_fluents)
                if not evaluate_expression(grounded_expression.root):
                    self.logger.debug(f"The evaluation of the numeric state variable failed. "
                                      f"The failed expression:\n{str(grounded_expression)}")
                    return False

            except KeyError:
                return False

        return True

    @staticmethod
    def _group_effect_predicates(grounded_effects: Set[GroundedPredicate]) -> Dict[str, Set[GroundedPredicate]]:
        """

        :param grounded_effects:
        :return:
        """
        grouped_effects = defaultdict(set)
        for predicate in grounded_effects:
            grouped_effects[predicate.lifted_untyped_representation].add(predicate)

        return grouped_effects

    @staticmethod
    def _update_single_numeric_expression(numeric_expression: NumericalExpressionTree,
                                          previous_values: Dict[str, PDDLFunction]) -> None:
        """Updates the numeric value of a single numeric expression.

        :param numeric_expression: the expression that represents the change to the state.
        :param previous_values: the previous values of the numeric expressions in the state.
        """
        set_expression_value(numeric_expression.root, previous_values)
        new_grounded_function = evaluate_expression(numeric_expression.root)
        previous_values[new_grounded_function.untyped_representation] = new_grounded_function

    def _validate_antecedents_hold(self, state: State, conditional_effect: ConditionalEffect) -> bool:
        """Validates whether the antecedents conditions hold for a conditional effect.

        :param state: the state the effect is applied on.
        :param conditional_effect: the conditional effect to test.
        :return: whether the antecedent conditions hold.
        """
        return (self._positive_preconditions_hold(state, conditional_effect.positive_conditions)
                and self._negative_preconditions_hold(state, conditional_effect.negative_conditions) and
                self._numeric_conditions_set_hold(state, conditional_effect.numeric_conditions))

    def _update_delete_effects(self, next_state_predicates: Dict[str, Set[GroundedPredicate]]) -> None:
        """Updates the state with the delete effects.

        :param next_state_predicates: the next state predicates.
        """
        grouped_delete_effects = self._group_effect_predicates(self.grounded_delete_effects)
        self.logger.debug("Removing state predicates according to the delete effects.")
        for lifted_predicate_str, grounded_predicates in grouped_delete_effects.items():
            next_state_grounded_predicates = next_state_predicates[lifted_predicate_str]
            if len(next_state_grounded_predicates) == 0:
                next_state_predicates[lifted_predicate_str] = set()
                continue

            for predicate_to_remove in grounded_predicates:
                for next_state_predicate in next_state_grounded_predicates:
                    if predicate_to_remove.untyped_representation == next_state_predicate.untyped_representation:
                        next_state_grounded_predicates.discard(next_state_predicate)
                        break

            next_state_predicates[lifted_predicate_str] = next_state_grounded_predicates

    def _update_add_effects(self, next_state_predicates: Dict[str, Set[GroundedPredicate]]) -> None:
        """Updates the state with the add effects.

        :param next_state_predicates: the next state predicates.
        """
        grouped_add_effects = self._group_effect_predicates(self.grounded_add_effects)
        self.logger.debug("Adding the new predicates according to the add effects.")
        for lifted_predicate_str, grounded_predicates in grouped_add_effects.items():
            updated_predicates = next_state_predicates[lifted_predicate_str].union(grounded_predicates)
            next_state_predicates[lifted_predicate_str] = updated_predicates

    def update_state_predicates(self, previous_state: State) -> Dict[str, Set[GroundedPredicate]]:
        """Updates the state predicates based on the action that is being applied.

        :param previous_state: the state that the action is being applied on.
        :return: a set of predicates representing the next state.
        """
        self.logger.info("Applying the action on the state predicates.")
        next_state_predicates = defaultdict(set)
        for lifted_predicate_name, grounded_predicates in previous_state.state_predicates.items():
            next_state_predicates[lifted_predicate_name] = \
                set([GroundedPredicate(p.name, p.signature, p.object_mapping) for p in grounded_predicates])

        self._update_delete_effects(next_state_predicates)
        self._update_add_effects(next_state_predicates)

        self.logger.debug("Applying the conditional discrete effects!")
        self.update_discrete_conditional_effects(previous_state, next_state_predicates,
                                                 self.grounded_conditional_effects)
        self.logger.debug("Applying universal effects on the state!")
        self.update_universal_effects(previous_state, next_state_predicates)

        return next_state_predicates

    def update_universal_effects(
            self, previous_state: State, next_state_predicates: Dict[str, Set[GroundedPredicate]]) -> None:
        """Updates the state predicates based on the universal effects of the action.

        :param previous_state: the state that the action is being applied on.
        :param next_state_predicates: the next state predicates.
        """
        if self.problem_objects is None:
            self.logger.warning("Did not receive the problem object so cannot apply the universal effects.")
            return

        for universal_effect in self.lifted_universal_effects:
            self.logger.debug("Updating the action's signature to temporarily include the quantified parameter.")
            self.action.signature[universal_effect.quantified_parameter] = universal_effect.quantified_type
            for pddl_object in self.problem_objects.values():
                if pddl_object.type.name != universal_effect.quantified_type.name:
                    continue

                self.logger.debug(f"Trying to apply the universal effect on the object: {str(pddl_object)}")
                extended_parameter_map = {lifted_param: grounded_object
                                          for lifted_param, grounded_object in
                                          zip(self.action.signature, self.grounded_call_objects)}
                extended_parameter_map[universal_effect.quantified_parameter] = pddl_object.name
                grounded_conditional_effect = self._ground_conditional_effect(universal_effect.conditional_effect,
                                                                              extended_parameter_map)
                self.update_discrete_conditional_effects(previous_state, next_state_predicates,
                                                         {grounded_conditional_effect})

            self.logger.debug("Removing the temporarily added signature item from the action.")
            self.action.signature.pop(universal_effect.quantified_parameter)

    def update_discrete_conditional_effects(
            self, previous_state: State, next_state_predicates: Dict[str, Set[GroundedPredicate]],
            conditional_effects: Set[ConditionalEffect]) -> None:
        """Checks whether the conditions for the conditional effects hold and updates the discrete state accordingly.

        :param previous_state: the state that the action is being applied on.
        :param next_state_predicates: the next state predicates.
        :param conditional_effects: the conditional effects that need to be applied.
        """
        for effect in conditional_effects:
            if not self._validate_antecedents_hold(previous_state, effect):
                self.logger.debug(
                    f"Some of the antecedents for the conditional effect do not hold for the action {self.name}.")
                continue

            self.logger.debug("The antecedents for the effect hold so applying the effect.")
            for predicate in effect.add_effects:
                lifted_predicate_str = predicate.lifted_untyped_representation
                next_state_grounded_predicates = next_state_predicates.get(lifted_predicate_str, set())
                next_state_grounded_predicates.add(predicate)
                next_state_predicates[lifted_predicate_str] = next_state_grounded_predicates

            for predicate in effect.delete_effects:
                next_state_predicates[predicate.lifted_untyped_representation].discard(predicate)

    def update_numeric_conditional_effects(
            self, previous_state: State, new_state_numeric_fluents: Dict[str, PDDLFunction]) -> None:
        """Checks whether the conditions for the conditional effects hold and updates the discrete state accordingly.

        :param previous_state: the state that the action is being applied on.
        :param new_state_numeric_fluents: the next state grounded numeric functions.
        """
        for effect in self.grounded_conditional_effects:
            if not (self._positive_preconditions_hold(previous_state, effect.positive_conditions)
                    and self._negative_preconditions_hold(previous_state, effect.negative_conditions) and
                    self._numeric_conditions_set_hold(previous_state, effect.numeric_conditions)):
                continue

            self.logger.debug("The conditionals for the effect hold so applying the numeric effect.")
            for grounded_expression in effect.numeric_effects:
                self._update_single_numeric_expression(grounded_expression, new_state_numeric_fluents)

    def update_state_functions(self, previous_state: State) -> Dict[str, PDDLFunction]:
        """Updates the state functions based on the action that is being applied.

        :param previous_state: the state that the action is being applied on.
        :return: the updated state functions with their new values.
        """
        new_state_numeric_fluents = {**previous_state.state_fluents}
        for grounded_expression in self.grounded_numeric_effects:
            try:
                self.logger.debug("First setting the values of the functions according to the stored state.")
                self._update_single_numeric_expression(grounded_expression, new_state_numeric_fluents)

            except KeyError:
                raise ValueError(f"There are missing fluents in the state! "
                                 f"State fluents: {previous_state.state_fluents}\n"
                                 f"Requested grounded expression:\n {str(grounded_expression)}")

        return new_state_numeric_fluents

    def is_applicable(self, state: State) -> bool:
        """Checks if the action is applicable on the current state.

        :param state: the state prior to the action's execution.
        :return: whether the action is applicable.
        """
        if not self.grounded:
            self.ground()

        if not self._positive_preconditions_hold(state) or not self._negative_preconditions_hold(state):
            return False

        # Checking for objects equality.
        if len(self.grounded_equality_preconditions) > 0 and \
                not self._equality_holds(self.grounded_equality_preconditions) or \
                len(self.grounded_inequality_preconditions) > 0 and \
                self._equality_holds(self.grounded_inequality_preconditions):
            return False

        # Checking that the value of the numeric expression holds.
        if not self._numeric_conditions_set_hold(state, self.grounded_numeric_preconditions):
            return False

        if len(self.grounded_disjunctive_numeric_preconditions) > 0 and \
                not any([self._numeric_conditions_set_hold(state, disjunctive_numeric_preconditions)
                         for disjunctive_numeric_preconditions in self.grounded_disjunctive_numeric_preconditions]):
            self.logger.debug("None of the disjunctive numeric preconditions hold.")
            return False

        return True

    def apply(self, previous_state: State, allow_inapplicable_actions: bool = False) -> State:
        """Applies an action on a state and changes the state according to the action's effects.

        :param previous_state: the state in which the operator is being applied on.
        :param allow_inapplicable_actions: whether to allow inapplicable actions to be applied.
        :return: the new state that was created by applying the operator.
        """
        # First need to apply the operator's discrete effects.
        if not self.grounded:
            self.ground()

        if not self.is_applicable(previous_state):
            self.logger.warning("Tried to apply an action to a state where the action's preconditions don't hold!")
            if not allow_inapplicable_actions:
                raise ValueError()

        self.logger.debug(f"Applying the grounded action - {self.name} on the current state.")
        next_state_predicates = self.update_state_predicates(previous_state)
        next_state_numeric_fluents = self.update_state_functions(previous_state)
        return State(predicates=next_state_predicates, fluents=next_state_numeric_fluents)

    def ground(self) -> None:
        """grounds the operator's preconditions and effects."""
        # First matching the lifted action signature to the grounded objects.
        parameters_map = {lifted_param: grounded_object
                          for lifted_param, grounded_object in zip(self.action.signature, self.grounded_call_objects)}

        self.grounded_positive_preconditions = self._ground_predicates(self.action.positive_preconditions,
                                                                       parameters_map)
        self.grounded_negative_preconditions = self._ground_predicates(self.action.negative_preconditions,
                                                                       parameters_map)
        self.grounded_add_effects = self._ground_predicates(self.action.add_effects, parameters_map)
        self.grounded_delete_effects = self._ground_predicates(self.action.delete_effects, parameters_map)
        self.grounded_equality_preconditions = self._ground_equality_objects(self.action.equality_preconditions,
                                                                             parameters_map)
        self.grounded_inequality_preconditions = self._ground_equality_objects(self.action.inequality_preconditions,
                                                                               parameters_map)

        self.grounded_numeric_preconditions = self._ground_numeric_expressions(self.action.numeric_preconditions,
                                                                               parameters_map)
        self.grounded_disjunctive_numeric_preconditions = [
            self._ground_numeric_expressions(expressions, parameters_map) for expressions in
            self.action.disjunctive_numeric_preconditions]
        self.grounded_numeric_effects = self._ground_numeric_expressions(self.action.numeric_effects, parameters_map)
        self.grounded_conditional_effects = {self._ground_conditional_effect(condition, parameters_map) for condition in
                                             self.action.conditional_effects}
        self.lifted_universal_effects = self.action.universal_effects
        self.grounded = True


class NOPOperator:
    """A no-op operator for when agents do not take action in a certain timestamp."""

    def __str__(self):
        return "(nop )"

    @property
    def name(self):
        return "nop"



"""Parse the output of the mtericFF planner so that the algorithms will be able to use it."""
import logging
import re
import sys

from pathlib import Path
from typing import NoReturn, List, Tuple

PLAN_COMPONENT_REGEX = r"\d: ([\w+\s?-]+)\n"
VALID_PLAN_FOUND_PATTERN = "ff: found legal plan as follows"
NO_SOLUTION_OPTIONS = ["problem proven unsolvable.",
                       "ff: goal can be simplified to FALSE. No plan will solve it",
                       "all increasers applied yet goal not fulfilled"]
NO_SOLUTION_FOUND_PATTERN = "problem proven unsolvable."
NO_SOLUTION_FOUND_PATTERN_2 = "ff: goal can be simplified to FALSE. No plan will solve it"
NO_SOLUTION_FOUND_PATTERN_3 = "all increasers applied yet goal not fulfilled"


class MetricFFParser:
    """Parse metricFF plans and exports then into standard output file."""

    logger: logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @staticmethod
    def _open_plan_file(input_path: Path) -> str:
        """Safely open the file and returns its content.

        :param input_path: the path to the log file.
        :return: the file content.
        """
        with open(input_path, "rt") as interm_plan_file:
            return interm_plan_file.read()

    def _parse_plan_content(self, planner_output) -> List[str]:
        """Parse the content of the file and export the action sequence.

        :param planner_output: the file content with the possible action sequence.
        :return: the action sequence.
        """
        matches = re.finditer(PLAN_COMPONENT_REGEX, planner_output, re.MULTILINE)
        plan_seq = []
        for match in matches:
            action_sequence = match.group(1)
            self.logger.debug(f"action sequence - {action_sequence}")
            plan_seq.append(f"({action_sequence.lower().strip()})\n")

        if len(plan_seq) == 0:
            return []

        return plan_seq

    def parse_plan(self, input_path: Path, output_path: Path) -> NoReturn:
        """Parse the output file and exports a plan if exists.

        :param input_path: the path to the output log of metricFF planner.
        :param output_path: the path to the output plan file.
        """
        planner_output = self._open_plan_file(input_path)
        action_sequence = self._parse_plan_content(planner_output)
        if len(action_sequence) == 0:
            return

        with open(output_path, "wt") as output_file:
            output_file.writelines(action_sequence)

    def get_solving_status(self, input_path: Path) -> Tuple[str, List[str]]:
        """Validates if the plan file contains a valid plan.

        :param input_path: the path to the input file.
        :return: "ok" if a plan exists, "timeout" if the solver timed out or "no-solution"
            if there is no solution for the problem as well as the action sequence.
        """
        file_content = self._open_plan_file(input_path)
        plan_found = re.search(VALID_PLAN_FOUND_PATTERN, file_content, re.MULTILINE)
        if plan_found is not None:
            action_sequence = self._parse_plan_content(file_content)
            return "ok", action_sequence

        for option in NO_SOLUTION_OPTIONS:
            no_solution_match = re.search(option, file_content, re.MULTILINE)
            if no_solution_match is not None:
                return "no-solution", []

        return "timeout", []


if __name__ == '__main__':
    MetricFFParser().parse_plan(Path(sys.argv[1]), Path(sys.argv[2]))



"""Module to convert MA domains into single agent domains"""
import logging
from pathlib import Path

from pddl_plus_parser.exporters import DomainExporter
from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Domain


class MultiAgentDomainsConverter:
    """Converts multiple MA domains to single agent domain containing all the agents' data."""

    logger: logging.Logger
    domains_directory_path: Path

    def __init__(self, working_directory_path: Path):
        self.logger = logging.getLogger(__name__)
        self.domains_directory_path = working_directory_path

    def locate_domains(self) -> Domain:
        """Locate only the domains when there is no need to also parse the problems."""
        combined_domain = Domain()
        for domain_path in self.domains_directory_path.glob("domain-*.pddl"):
            self.logger.debug(f"Working on the domain - {domain_path.stem}")
            domain_file_name = domain_path.stem
            agent_domain = DomainParser(domain_path=domain_path, partial_parsing=False,
                                        enable_disjunctions=True).parse_domain()

            combined_domain.name = agent_domain.name
            combined_domain.requirements = agent_domain.requirements
            combined_domain.types.update(agent_domain.types)
            combined_domain.predicates.update(agent_domain.predicates)
            combined_domain.constants.update(agent_domain.constants)
            combined_domain.actions.update(agent_domain.actions)
            combined_domain.functions.update(agent_domain.functions)

            self.logger.debug(f"extracted the domain - {domain_file_name}")

        return combined_domain

    def export_combined_domain(self) -> None:
        """Export the multi-agent domains to a single PDDL domain file."""
        combined_domain = self.locate_domains()
        DomainExporter().export_domain(combined_domain, self.domains_directory_path / f"combined_domain.pddl")



import os
from pathlib import Path

CWD = os.getcwd()
TEST_PARSING_FILE_PATH = Path(CWD, "test_domain_format_file.pddl")
TEST_WOODWORKING_DOMAIN_PATH = Path(CWD, "woodworking_domain.pddl")
TEST_NUMERIC_DEPOT_DOMAIN_PATH = Path(CWD, "depot_numeric_domain.pddl")
TEST_NUMERIC_PROBLEM = Path(CWD, "test_agricola_problem.pddl")
TEST_NUMERIC_DOMAIN = Path(CWD, "test_agricola_domain.pddl")

TEST_NUMERIC_DEPOT_DOMAIN = Path(CWD, "depot_numeric.pddl")
TEST_NUMERIC_DEPOT_PROBLEM = Path(CWD, "pfile2.pddl")
TEST_NUMERIC_DEPOT_TRAJECTORY = Path(CWD, "test_numeric_trajectory")

FARMLAND_NUMERIC_DOMAIN = Path(CWD, "farmland.pddl")
FARMLAND_NUMERIC_PROBLEM = Path(CWD, "pfile10_10.pddl")
FARMLAND_NUMERIC_TRAJECTORY = Path(CWD, "pfile10_10.trajectory")

ZENOTRAVEL_DOMAIN_PATH = Path(CWD, "zenonumeric.pddl")
ZENOTRAVEL_PROBLEM_PATH = Path(CWD, "pfile0.pddl")

PLANT_WATERING_DOMAIN = Path(CWD, "plant_watering_domain.pddl")

WOODWORKING_COMBINED_DOMAIN_PATH = Path(CWD) / "woodworking_combined_domain.pddl"
WOODWORKING_COMBINED_PROBLEM_PATH = Path(CWD) / "woodworking_combined_problem.pddl"
WOODWORKING_COMBINED_TRAJECTORY_PATH = Path(CWD) / "ma_woodworking_trajectory.trajectory"


TEST_PREDICATES_FOR_CONDITIONAL_DOMAIN = """((on ?c1 - card ?c2 - cardposition)
    (clear ?c - cardposition)
    (in-play ?c - card)
    (current-deal ?d - deal)
    (CAN-CONTINUE-GROUP ?c1 - card ?c2 - cardposition)
    (CAN-BE-PLACED-ON ?c1 - card ?c2 - card)
    (IS-ACE ?c - card)
    (IS-KING ?c - card)
    (NEXT-DEAL ?d - deal ?nd - deal)
    (TO-DEAL ?c - card ?p - tableau ?d - deal ?next - cardposition)
    (currently-dealing)
    (currently-collecting-deck)
    (collect-card ?c - cardposition)
    (part-of-tableau ?c - cardposition ?t - tableau)
    (movable ?c - card)
    (currently-updating-unmovable)
    (make-unmovable ?c - card)
    (currently-updating-movable)
    (make-movable ?c - cardposition)
    (currently-updating-part-of-tableau)
    (make-part-of-tableau ?c - card ?t - tableau)
)
"""

TEST_PREDICATES_FOR_UNIVERSAL_QUANTIFIER_DOMAIN = """(
    (origin ?person - passenger ?floor - floor)
    (destin ?person - passenger ?floor - floor)
    (above ?floor1 - floor  ?floor2 - floor)
    (boarded ?person - passenger)
    (served ?person - passenger)
    (lift-at ?floor - floor)
)
"""


TEST_TYPES_FOR_CONDITIONAL_DOMAIN = """(cardposition - object
    card_or_tableau - cardposition
    card - card_or_tableau
    tableau - card_or_tableau
    deal - cardposition
)"""

TYPES_FOR_UNIVERSAL_CONDITIONAL_DOMAIN = """(passenger - object
          floor - object)"""

TEST_CONSTANTS_FOR_CONDITIONAL_DOMAIN = """(discard - cardposition)"""

SPIDER_DOMAIN_PATH = Path(CWD, "domain_spider.pddl")
SPIDER_PROBLEM_PATH = Path(CWD, "pfile04.pddl")

DEPOT_MA_DOMAIN_PATH = Path(CWD, "Depots.pddl")
DEPOT_MA_PROBLEM_PATH = Path(CWD, "pfile1_depot.pddl")
DEPOT_MA_TRAJECTORY_PATH = Path(CWD, "test_joint_trajectory_with_potential_bug")


"""Module containing utility functionality for the parsing process."""
from collections import Iterator
from typing import Dict

from pddl_plus_parser.models import PDDLType, SignatureType


def parse_signature(parameters: Iterator[str], domain_types: Dict[str, PDDLType]) -> SignatureType:
    """Parse the signature of a statement.

    :param parameters: the parameters that appear in the signature.
    :param domain_types: the types that were extracted from the domain.
    :return: the object representing the signature's data.
    """
    signature = {}
    for parameter_name in parameters:
        # now the next item is the dash that we need to ignore
        next(parameters)
        parameter_type = next(parameters)
        signature[parameter_name] = domain_types[parameter_type]

    return signature


from collections import defaultdict
from typing import List

from pddl_plus_parser.models import Problem, State, ActionCall, Domain, Operator, GroundedPredicate, \
    NOP_ACTION


def create_initial_state(problem: Problem) -> State:
    """Create the initial state of the problem.

    :param problem: the problem object.
    :return: the initial state of the problem.
    """
    initial_state_predicates = problem.initial_state_predicates
    initial_state_numeric_fluents = problem.initial_state_fluents
    return State(predicates=initial_state_predicates, fluents=initial_state_numeric_fluents, is_init=True)


def apply_actions(domain: Domain, current_state: State, joint_action: List[ActionCall]) -> State:
    """

    :param domain: the domain with the action scheme.
    :param current_state: the current state that the action is being applied on.
    :param joint_action: the executable actions of the agents.
    :return: The state resulting from applying the actions.
    """
    operators = []
    if len(joint_action) == 1:
        action_call = joint_action[0]
        action = domain.actions[action_call.name]
        return Operator(action=action, domain=domain, grounded_action_call=action_call.parameters).apply(current_state)

    accumulative_discrete_effects = defaultdict(set)
    accumulative_numeric_effects = current_state.state_fluents.copy()
    for lifted_name, grounded_predicates in current_state.state_predicates.items():
        for predicate in grounded_predicates:
            accumulative_discrete_effects[lifted_name].add(GroundedPredicate(predicate.name, predicate.signature,
                                                                             predicate.object_mapping))

    for action_call in joint_action:
        if action_call.name == NOP_ACTION:
            continue

        action = domain.actions[action_call.name]
        operator = Operator(action=action, domain=domain, grounded_action_call=action_call.parameters)
        partial_numeric_state = State(predicates=current_state.state_predicates, fluents=accumulative_numeric_effects)
        partial_next_state = operator.apply(partial_numeric_state)
        accumulative_numeric_effects.update(partial_next_state.state_fluents)
        # since there are multiple actions being executed at once, we need to take the difference between the
        # current state and the next state and accumulate it.
        for delete_effect in operator.grounded_delete_effects:
            for grounded_predicate in accumulative_discrete_effects[delete_effect.lifted_untyped_representation]:
                if grounded_predicate.untyped_representation == delete_effect.untyped_representation:
                    accumulative_discrete_effects[delete_effect.lifted_untyped_representation].remove(
                        grounded_predicate)
                    break

        for add_effect in operator.grounded_add_effects:
            accumulative_discrete_effects[add_effect.lifted_untyped_representation].add(add_effect)

        operators.append(operator)

    return State(predicates=accumulative_discrete_effects, fluents=accumulative_numeric_effects)



"""Module that represents a state definition in a PDDL trajectory."""
from typing import Dict, Set

from .pddl_function import PDDLFunction
from .pddl_predicate import GroundedPredicate


class State:
    """A representation of a state in a trajectory."""

    is_init: bool
    # Maps between a lifted predicate definition to all of its problem groundings
    state_predicates: Dict[str, Set[GroundedPredicate]]
    # Map between the grounded numeric fluent string to the actual function.
    state_fluents: Dict[str, PDDLFunction]

    def __init__(self, predicates: Dict[str, Set[GroundedPredicate]],
                 fluents: Dict[str, PDDLFunction], is_init: bool = False):
        self.state_predicates = predicates
        self.state_fluents = fluents
        self.is_init = is_init

    def _serialize_numeric_fluents(self) -> str:
        """Serialize the numeric fluents of the state.

        :return: the string representing the assigned grounded fluents.
        """
        return " ".join(fluent.state_representation for fluent in self.state_fluents.values())

    def _serialize_predicates(self) -> str:
        """Serialize the predicates the constitute the state's facts.

        :return: the string representation of the state's facts. 
        """
        predicates_str = ""
        for grounded_predicates in self.state_predicates.values():
            predicates_str += " "
            predicates_str += " ".join(predicate.untyped_representation for predicate in grounded_predicates)

        return predicates_str

    def typed_serialize(self) -> str:
        """Returns a typed version of the serialized state.

        :return: typed version of the serialized state.
        """
        typed_predicates_str = ""
        for grounded_predicates in self.state_predicates.values():
            typed_predicates_str += " "
            typed_predicates_str += " ".join(str(predicate) for predicate in grounded_predicates)

        return f"({' '.join(fluent.state_typed_representation for fluent in self.state_fluents.values())}" \
               f"{typed_predicates_str})\n"

    def serialize(self) -> str:
        return f"({':init' if self.is_init else ':state'}" \
               f" {self._serialize_numeric_fluents()}" \
               f"{self._serialize_predicates()})\n"



from anytree import RenderTree, PreOrderIter, AnyNode

from pddl_plus_parser.models import construct_expression_tree, PDDLFunction, PDDLType, calculate, evaluate_expression, \
    NumericalExpressionTree

SIMPLE_EXPRESSION = ['assign', ['amount', '?jug1'], '0']
COMPLEX_EXPRESSION = ['>=', ['-', ['capacity', '?jug2'], ['amount', '?jug2']], ['amount', '?jug1']]

TEST_DOMAIN_FUNCTIONS = {
    "capacity": PDDLFunction(name="capacity", signature={"?jug": PDDLType(name="jug")}),
    "amount": PDDLFunction(name="amount", signature={"?jug": PDDLType(name="jug")})
}


def test_construct_expression_tree_constructs_simple_expression_correctly():
    node_tree = construct_expression_tree(SIMPLE_EXPRESSION, TEST_DOMAIN_FUNCTIONS)
    print(RenderTree(node_tree))

    expected_node_ids = ["assign", "(amount ?jug1 - jug)", "0.0"]
    for index, node in enumerate(PreOrderIter(node_tree)):
        assert node.id == expected_node_ids[index]


def test_construct_expression_tree_constructs_complex_expression_correctly():
    node_tree = construct_expression_tree(COMPLEX_EXPRESSION, TEST_DOMAIN_FUNCTIONS)
    print(RenderTree(node_tree))

    expected_node_ids = [">=", "-", "(capacity ?jug2 - jug)", "(amount ?jug2 - jug)", "(amount ?jug1 - jug)"]
    for index, node in enumerate(PreOrderIter(node_tree)):
        assert node.id == expected_node_ids[index]


def test_calculate_returns_correct_calculation_on_simple_tree():
    root = AnyNode(id="root", value="*", children=[
        AnyNode(id="sub0", value=12),
        AnyNode(id="sub1", value=5)
    ])
    assert calculate(root) == 60.0


def test_calculate_returns_correct_calculation_on_a_more_complex_tree():
    root = AnyNode(id="root", value="*", children=[
        AnyNode(id="d1", value="+", children=[
            AnyNode(id="sub0", value=2),
            AnyNode(id="sub1", value=5)
        ]),
        AnyNode(id="d2", value="-", children=[
            AnyNode(id="d3", value="/", children=[
                AnyNode(id="sub0", value=100),
                AnyNode(id="sub1", value=4)
            ]),
            AnyNode(id="sub4", value=5)
        ])
    ])
    assert calculate(root) == 140.0


def test_evaluate_expression_evaluates_simple_pddl_ast():
    test_expression = ['>=', ['amount', '?jug1'], '-3.5']
    root = construct_expression_tree(test_expression, TEST_DOMAIN_FUNCTIONS)
    print(RenderTree(root))

    result = evaluate_expression(root)
    assert result == True


def test_evaluate_expression_evaluates_another_simple_pddl_ast():
    test_expression = ['>=', ['amount', '?jug1'], '3.5']
    root = construct_expression_tree(test_expression, TEST_DOMAIN_FUNCTIONS)
    print(RenderTree(root))

    result = evaluate_expression(root)
    assert result == False


def test_evaluate_expression_evaluates_complex_pddl_ast():
    test_expression = ['>=', ['/', ['+', ['amount', '?jug1'], '12'], ['+', ['amount', '?jug2'], '4']], '4.5']
    root = construct_expression_tree(test_expression, TEST_DOMAIN_FUNCTIONS)
    print(RenderTree(root))

    result = evaluate_expression(root)
    assert result == False


def test_convert_to_pddl_returns_correct_expression():
    test_expression = ['>=', ['capacity', '?jug2'], ['amount', '?jug2']]
    root = construct_expression_tree(test_expression, TEST_DOMAIN_FUNCTIONS)
    tree = NumericalExpressionTree(root)
    pddl_str = tree.to_pddl()
    assert pddl_str == "(>= (capacity ?jug2) (amount ?jug2))"



"""Exports a problem object to a PDDL file."""
from pathlib import Path
from typing import NoReturn, List, Dict, Set, Union

from pddl_plus_parser.models import PDDLObject, GroundedPredicate, Problem, PDDLFunction, NumericalExpressionTree


class ProblemExporter:
    """Class that is able to export a domain to a correct PDDL file."""

    @staticmethod
    def write_objects(problem_objects: Dict[str, PDDLObject]) -> str:
        """Writes the definitions of the types according to the PDDL file format.

        :param problem_objects: the objects that are available in the learned domain.
        :return: the formatted string representing the objects in the PDDL problem file.
        """
        objects = []
        for pddl_object in problem_objects.values():
            objects.append(str(pddl_object))

        return "(:objects\n{objects_content}\n)\n".format(objects_content="\n\t".join(objects))

    def write_initial_state(self, initial_state_predicates: Dict[str, Set[GroundedPredicate]],
                            initial_state_fluents: Dict[str, PDDLFunction]) -> str:
        """Writes the definitions of the types according to the PDDL file format.

        :param initial_state_predicates: the objects that are available in the learned domain.
        :return: the formatted string representing the state in the PDDL problem file.
        :param initial_state_fluents: the numeric fluents in the initial state.
        """
        predicates_str = []
        fluents_str = []
        for grounded_predicates_set in initial_state_predicates.values():
            predicates_str.append(self.extract_state_predicates(grounded_predicates_set))

        for fluent in initial_state_fluents.values():
            fluents_str.append(fluent.state_representation)

        joint_state_str = "\n\t".join([*predicates_str, *fluents_str])
        return f"(:init\n\t{joint_state_str}\n\n)\n"

    def write_goal_state(self, goal_state_predicates: List[GroundedPredicate],
                         goal_state_fluents: Set[NumericalExpressionTree]) -> str:
        """Writes the definitions of the types according to the PDDL file format.

        :param goal_state_predicates: the objects that are available in the learned domain.
        :return: the formatted string representing the state in the PDDL problem file.
        :param goal_state_fluents: the numeric expressions in the goal state.
        """
        predicates_str = self.extract_state_predicates(goal_state_predicates)
        goal_fluents = [fluent.to_pddl() for fluent in goal_state_fluents]

        joint_goal = "\n\t\t".join([predicates_str, *goal_fluents])
        return f"(:goal\n\t(and\n\t{joint_goal}\t\t\n)\n)\n"

    @staticmethod
    def extract_state_predicates(state: Union[List[GroundedPredicate], Set[GroundedPredicate]]) -> str:
        """Extract the needed problem predicates for the PDDL file representation.

        :param state: the state to write in a PDDL format.
        :return: the strings of containing the state's data.
        """
        return "\n\t".join([predicate.untyped_representation for predicate in state])

    def extract_problem(self, problem: Problem) -> str:
        """Extract the problem str from the problem object.

        :param problem: the problem object to extract the data to a PDDL string representation.
        """
        problem_objects = self.write_objects(problem.objects)
        initial_state = self.write_initial_state(problem.initial_state_predicates, problem.initial_state_fluents)
        goal_state = self.write_goal_state(problem.goal_state_predicates, problem.goal_state_fluents)
        problem_data = f"(define (problem {problem.name}) (:domain {problem.domain.name})\n" \
                       f"{problem_objects}\n" \
                       f"{initial_state}\n" \
                       f"{goal_state}\n)"

        return problem_data

    def export_problem(self, problem: Problem, export_path: Path) -> NoReturn:
        """Export the domain object to a correct PDDL file.

        :param problem: the problem object to export to a PDDL file.
        :param export_path: the path to the file that the domain would be exported to.
        """
        problem_data = self.extract_problem(problem)

        with open(export_path, "wt") as export_problem_file:
            export_problem_file.write(problem_data)



"""Module to represent PDDL problems and their data."""
from collections import defaultdict
from typing import List, Dict, Set

from .numerical_expression import NumericalExpressionTree
from .pddl_domain import Domain
from .pddl_function import PDDLFunction
from .pddl_object import PDDLObject
from .pddl_predicate import GroundedPredicate

METRICS = {
    "maximize": max,
    "minimize": min
}


class Problem:
    """Class representing a single PDDL problem."""

    domain: Domain
    name: str
    objects: Dict[str, PDDLObject]
    # Maps between a lifted predicate definition to all of its problem groundings
    initial_state_predicates: Dict[str, Set[GroundedPredicate]]
    # Map between the grounded numeric fluent string to the actual function.
    initial_state_fluents: Dict[str, PDDLFunction]
    # Since we don't need to optimize the search on the goal state, the goal state will remain as list.
    goal_state_predicates: List[GroundedPredicate]
    goal_state_fluents: Set[NumericalExpressionTree]
    metric: Dict[str, NumericalExpressionTree]

    def __init__(self, domain: Domain):
        self.name = ""
        self.domain = domain
        self.objects = {}
        self.initial_state_predicates = defaultdict(set)
        self.initial_state_fluents = {}
        self.goal_state_predicates = []
        self.goal_state_fluents = set()
        self.metric = {}

    def __str__(self):
        grounded_predicates = []
        for p in self.initial_state_predicates.values():
            grounded_predicates.extend(p)

        return f"<Problem - {self.name} of domain - {self.domain.name}.\n" \
               f"Initial state predicates - {[str(p) for p in grounded_predicates]}\n" \
               f"Initial state numeric fluents - {[f for f in self.initial_state_fluents]}\n" \
               f"Goal state - {[str(p) for p in self.goal_state_predicates]}" \
               ">"



"""Module test for the PDDLFunction class."""

from pddl_plus_parser.models import PDDLFunction, PDDLType

OBJECT_TYPE = PDDLType(name="object")
CITY_TYPE = PDDLType(name="city", parent=OBJECT_TYPE)


def test_state_representation_returns_correct_string_when_function_is_lifted():
    """Test that the state representation returns the correct string."""
    fluent = PDDLFunction(name="distance", signature={"?c1": CITY_TYPE, "?c2": CITY_TYPE})
    assert fluent.state_representation == "(= (distance ?c1 ?c2) 0)"


def test_state_representation_returns_correct_string_when_function_is_grounded_with_different_variables():
    """Test that the state representation returns the correct string."""
    fluent = PDDLFunction(name="distance", signature={"city0": CITY_TYPE, "city1": CITY_TYPE})
    assert fluent.state_representation == "(= (distance city0 city1) 0)"


def test_state_representation_returns_correct_string_when_function_is_grounded_with_the_same_variables():
    """Test that the state representation returns the correct string."""
    fluent = PDDLFunction(name="distance", signature={"city0": CITY_TYPE}, repeating_variables={"city0": 2})
    assert fluent.state_representation == "(= (distance city0 city0) 0)"


def test_state_typed_representation_returns_correct_string_when_function_is_grounded_with_the_same_variables():
    """Test that the state representation returns the correct string."""
    fluent = PDDLFunction(name="distance", signature={"city0": CITY_TYPE}, repeating_variables={"city0": 2})
    assert fluent.state_typed_representation == "(= (distance city0 - city city0 - city) 0)"



"""Module that encapsulates the multi-agent trajectory functionalities."""
import logging
import re
from pathlib import Path
from typing import List, Optional, Union

from pddl_plus_parser.models import Domain, Problem, Operator, State, JointActionCall, ActionCall, NOP_ACTION, \
    NOPOperator
from pddl_plus_parser.multi_agent.common import create_initial_state, apply_actions

JOINT_ACTION_REGEX = r"\(([\w+\s?-]+)\)"


def parse_action_call(joint_action_call: str) -> JointActionCall:
    """Parses the string representing the joint action call in the plan sequence.

    :param joint_action_call: the string representing the joint action.
    :return: the object representing the action name and its parameters.
    """
    matches = re.finditer(JOINT_ACTION_REGEX, joint_action_call)
    single_agent_actions = []
    for match in matches:
        single_agent_action_data = match.group(1)
        action_components = single_agent_action_data.split()
        action_name = action_components[0]
        single_agent_actions.append(ActionCall(name=action_name, grounded_parameters=action_components[1:]))

    return JointActionCall(single_agent_actions)


class MultiAgentTrajectoryTriplet:
    """Class representing a single multi-agent trajectory triplet."""
    previous_state: State
    joint_action: List[Union[Operator, NOPOperator]]
    next_state: State

    def __init__(self, previous_state: State, ops: List[Operator], next_state: State):
        self.previous_state = previous_state
        self.joint_action = ops
        self.next_state = next_state

    def __str__(self):
        return f"previous state: {self.previous_state.serialize()}\n" \
               f"operators: {[str(op) for op in self.joint_action]}\n" \
               f"next state: {self.next_state.serialize()}"


class MultiAgentTrajectoryExporter:
    """Export multi-agent trajectories in the appropriate format."""

    domain: Domain
    logger: logging.Logger

    def __init__(self, domain: Domain):
        self.domain = domain
        self.logger = logging.getLogger(__name__)

    def _read_plan(self, plan_file_path: Path) -> List[str]:
        """Read the plan file and exports the lines with the actions.

        :param plan_path: the path to the plan file.
        :return: the action sequence.
        """
        self.logger.debug(f"Reading the plan in the path {plan_file_path}")
        with open(plan_file_path, "rt") as plan_file:
            return plan_file.readlines()

    def create_multi_agent_triplet(self, previous_state: State, action_call: str) -> MultiAgentTrajectoryTriplet:
        """Create a single trajectory triplet by applying the joint action on the input state and combining the effects.

        :param previous_state: the state that the action is being applied on.
        :param action_call: the string representation of the grounded joint action call.
        :return: the new triplet containing (s,<a1, a2,..., am>,s').
        """
        self.logger.info(f"Trying to apply the action - {action_call} on the state - {previous_state.serialize()}")
        joint_action = parse_action_call(action_call)
        executed_actions = [ActionCall(name=op.name, grounded_parameters=op.parameters) for op in
                            joint_action.actions if op.name != NOP_ACTION]
        operators = []
        for sa_action in joint_action.actions:
            if sa_action.name == NOP_ACTION:
                operators.append(NOPOperator())
                continue

            operators.append(Operator(action=self.domain.actions[sa_action.name], domain=self.domain,
                                      grounded_action_call=sa_action.parameters))
        next_state = apply_actions(self.domain, previous_state, executed_actions)
        return MultiAgentTrajectoryTriplet(previous_state=previous_state, ops=operators, next_state=next_state)

    def parse_plan(self, problem: Problem, plan_path: Optional[Path] = None,
                   action_sequence: Optional[List[str]] = None) -> List[MultiAgentTrajectoryTriplet]:
        """Parse the input plan file to create the trajectory.

        :return: the list of triplets that was generated using the plan.
        """
        self.logger.info("Parsing the plan to extract the grounded operators.")
        plan_actions = action_sequence if action_sequence is not None else self._read_plan(plan_path)
        previous_state = create_initial_state(problem)
        triplets = []
        self.logger.debug("Starting to create the trajectory triplets.")
        for grounded_action_call in plan_actions:
            triplet = self.create_multi_agent_triplet(previous_state, grounded_action_call)
            triplets.append(triplet)
            previous_state = triplet.next_state

        return triplets

    @staticmethod
    def export(triplets: List[MultiAgentTrajectoryTriplet]) -> List[str]:
        """Export the input triplets as a valid trajectory object.

        :param triplets: the objects representing the triplets generated from the plan sequence.
        :return: a list of strings representing the trajectory.
        """
        serialized_trajectory = []
        first_state = triplets[0].previous_state
        serialized_trajectory.append(first_state.serialize())
        for triplet in triplets:
            operators = " ".join([str(op) for op in triplet.joint_action])
            serialized_trajectory.append(f"(operators: {operators})\n")
            serialized_trajectory.append(triplet.next_state.serialize())

        serialized_trajectory[0] = f"({serialized_trajectory[0]}"
        serialized_trajectory[-1] = f"{serialized_trajectory[-1]})"
        return serialized_trajectory

    def export_to_file(self, triplets: List[MultiAgentTrajectoryTriplet], output_path: Path) -> None:
        """Export the trajectory to a file.

        :param triplets: the trajectory triples.
        :param output_path: the path to the output file.
        """
        trajectory_lines = self.export(triplets)
        with open(output_path, "wt") as output_path:
            output_path.writelines(trajectory_lines)



"""Problem generator for the zeno-travel domain."""
import argparse
import random
from enum import Enum
from typing import List, Optional, Any, Dict

import numpy

MAX_RAND = 10 ** 6
random.seed(42)


class AllowedDomainTypes(Enum):
    """Enum for the allowed domain types."""
    STRIPS = 1
    NUMERIC = 2


class CityMap:
    """Defines the map of the problem."""
    num_locations: int
    num_distances: int
    map: numpy.ndarray
    domain_type: AllowedDomainTypes

    def __init__(self, num_locations: int, num_distances: int, domain_type: AllowedDomainTypes):
        self.domain_type = domain_type
        self.num_locations = num_locations
        self.num_distances = num_distances
        self.map = numpy.zeros((num_locations, num_distances))
        for i in range(num_locations):
            for j in range(i + 1, num_locations):
                self.map[i][j] = random.randint(0, num_distances // 2) + num_distances / 2
                self.map[j][i] = self.map[i][j]

    def define_city_map(self) -> List[str]:
        """

        :return:
        """
        map_str = []
        if self.domain_type == AllowedDomainTypes.NUMERIC:
            for i in range(self.num_locations):
                for j in range(self.num_locations):
                    map_str.append(f"\t(= (distance city{i} city{j}) {int(self.map[i][j])})")

        return map_str

    def define_city_objects(self) -> List[str]:
        """

        :return:
        """
        return [f"\tcity{i} - city" for i in range(self.num_locations)]


class Locatable:
    location: int
    destination: int
    interesting: bool
    id: int

    def __init__(self, num_location: int):
        self.location = random.randint(0, num_location)
        self.destination = random.randint(0, num_location)
        self.interesting = True


class Airplane(Locatable):
    """Defines the airplane."""
    id: int
    airplane_data: Dict[str, Any]
    domain_type: AllowedDomainTypes
    num_planes: int

    def __init__(self, id: int, num_locations: int, num_distances: int, num_planes: int,
                 domain_type: AllowedDomainTypes):
        super(Airplane, self).__init__(num_locations)
        self.id = id
        self.num_planes = num_planes
        slow_burn_rate = random.randint(1, 5)
        slow_speed = random.randint(0, 100) + 100
        self.airplane_data = {
            "slow_burn_rate": random.randint(1, 5),
            # "slow_speed": random.randint(0, 100) + 100,
            "fuel": random.randint(0, slow_burn_rate * num_distances),
            "capacity": random.randint(0, int(2.1 * random.random()) * slow_burn_rate * num_distances),
            # "fast_speed": random.randint(0, int(1.0 + random.random() * 2) * slow_speed),
            "fast_burn_rate": random.randint(0, int(2.0 + random.random() * 2) * slow_burn_rate),
            "refuel_rate": 2 * random.randint(0, slow_burn_rate * num_distances),
            "zoom_limit": 1 + random.randint(0, 10)
        }
        self.domain_type = domain_type
        if random.randint(0, 10) < 7:
            self.interesting = False

    def define_airplane_object(self) -> str:
        """

        :return:
        """
        return f"\tplane{self.id} - aircraft"

    def define_airplane_map(self) -> List[str]:
        """"""
        airplane_map = [f"\t(at plane{self.id} city{self.location})"]
        if self.domain_type == AllowedDomainTypes.NUMERIC:
            for field_name, field_value in self.airplane_data.items():
                airplane_map.append(f"\t(= ({field_name} plane{self.id}) {field_value})")

            airplane_map.append(f"\t(= (onboard plane{self.id}) 0)")

        return airplane_map

    def define_airplane_goal(self) -> Optional[str]:
        """

        :return:
        """
        if self.interesting:
            return f"\t(at plane{self.id} city{self.destination})"

        return None


class Person(Locatable):
    """Defines the person."""
    id: int
    person_data: Dict[str, Any]
    domain_type: AllowedDomainTypes
    num_people: int

    def __init__(self, id: int, num_people: int, domain_type: AllowedDomainTypes):
        super(Person, self).__init__(num_people)
        self.id = id
        self.num_people = num_people
        self.domain_type = domain_type
        if random.randint(0, 100) < 3:
            self.interesting = False

    def define_person_object(self) -> str:
        """

        :return:
        """
        return f"\tperson{self.id} - person"

    def define_person_map(self) -> str:
        """"""
        return f"\t(at person{self.id} city{self.location})"

    def define_person_goal(self) -> Optional[str]:
        """

        :return:
        """
        if self.interesting:
            return f"\t(at person{self.id} city{self.destination})"

        return None


class ZenoTravelGenerator:
    """Problem generator for the zeno-travel domain."""

    num_people: int
    num_planes: int
    people: List[Person]
    airplanes: List[Airplane]
    cities_map: CityMap
    domain_type: AllowedDomainTypes

    def __init__(self, domain_type: AllowedDomainTypes, number_people: int, number_airplanes: int, num_locations: int):
        """Initialize the generator."""
        self.num_people = number_people
        self.num_planes = number_airplanes
        self.domain_type = domain_type
        self.cities_map = CityMap(num_locations=num_locations, num_distances=100, domain_type=domain_type)
        self.people = [Person(id=index, num_people=number_people, domain_type=domain_type) for index in
                       range(number_people)]
        self.airplanes = [Airplane(id=index, num_locations=num_locations, num_distances=100, domain_type=domain_type,
                                   num_planes=number_airplanes) for index in range(number_airplanes)]

    def generate_problem(self) -> str:
        """Generate the problem."""
        problem = [f"(define (problem ZTRAVEL-{self.num_people}-{self.num_planes})",
                   "\t(:domain zeno-travel)", "(:objects"]
        for person in self.people:
            problem.append(person.define_person_object())

        for airplane in self.airplanes:
            problem.append(airplane.define_airplane_object())

        problem.extend(self.cities_map.define_city_objects())
        problem.append("\t)\n(:init")
        if self.domain_type == AllowedDomainTypes.STRIPS:
            for i in range(7):
                problem.append(f"\tfl{i} - flevel")

        for airplane in self.airplanes:
            problem.extend(airplane.define_airplane_map())

        for person in self.people:
            problem.append(person.define_person_map())

        problem.extend(self.cities_map.define_city_map())
        if self.domain_type == AllowedDomainTypes.NUMERIC:
            problem.append("\t(= (total-fuel-used) 0)")

        problem.append(")\n(:goal (and")
        for airplane in self.airplanes:
            goal = airplane.define_airplane_goal()
            if goal is not None:
                problem.append(goal)

        for person in self.people:
            goal = person.define_person_goal()
            if goal is not None:
                problem.append(goal)

        problem.append("\t))\n")
        if self.domain_type == AllowedDomainTypes.NUMERIC:
            # define the metric function
            value1 = random.randint(0, 5) + 1
            value2 = random.randint(0, 5) + 1
            problem.append(f"(:metric minimize (+ (* {value2} (total-time))  (* {value1} (total-fuel-used))))")

        problem.append(")")

        return "\n".join(problem)


def parse_arguments() -> argparse.Namespace:
    """Parse the command line arguments."""
    parser = argparse.ArgumentParser(description="Generate a single zenotravel planning problem.")
    parser.add_argument("--number_people", required=True, help="The number of people in the problem.", type=int)
    parser.add_argument("--number_airplanes", required=True, help="The number of airplanes in the problem.", type=int)
    parser.add_argument("--num_locations", required=True, help="The number of locations in the problem.", type=int)
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_arguments()
    generator = ZenoTravelGenerator(domain_type=AllowedDomainTypes.NUMERIC,
                                    number_people=args.number_people,
                                    number_airplanes=args.number_airplanes,
                                    num_locations=args.num_locations)
    print(generator.generate_problem())
